%IVDragTimer = 102

' Bare metal (won't even draw...)
Function NewIVRowCol( win As iWin ) Common Export As iExtend
	Local obj As iExtend
	obj = Class "cIVRowCol"
	obj.Extend( win )
	Function = obj
End Function

Class cIVRowCol Guid$( "{8bf1512e-32dc-4408-86c3-6a7d71487aaf}" ) Common
	Instance hLayout As iLayout: ' Horizontal layout
	Instance vLayout As iLayout: ' Vertical layout

	Instance dataSource_ As iDataSource

	Instance hNorm As Dword: ' Normal cursor
	Instance hHorz As Dword: ' Horizontal cursor
	Instance hVert As Dword: ' Vertical cursor
	Instance hBoth As Dword: ' Both cursor

	Instance hWnd_ As Dword:  ' Window handle

	Instance extensions_ As iExtensionList

	Instance copyHeaders_ As Long: ' 1 = 1st line is headers, 0 = No headers

	Instance dragging_ As Long: ' 0 = not dragging, <>0 dragging
	Instance dragX_ As Long:    ' Initial drag point
	Instance dragY_ As Long:    ' Initial drag point

	Instance sizing_ As Long:   ' Currently sizing
	Instance sizeX_ As Long:    ' x adjustment ( drift from right edge )
	Instance sizeY_ As Long:    ' y adjustment ( drift from bottom edge )
	Instance sizeR_ As Long:    ' Row offset of sizing row
	Instance sizeC_ As Long:    ' Col offset of sizing column

	Instance allowResize_ As Long: ' Types of resize allowed ( %IVResize )

	Instance drawVGrid_ As iRender
	Instance drawHGrid_ As iRender

	Instance drawContentBack_ As iRender: ' Fill portion with content in it.
	Instance drawControlBack_ As iRender: ' Fill control background

	Instance base_ As iIVBase
	Instance table_ As iIVRowCol

	Class Method Create()
		Local no As iRender
		Local i As Long

		no = Nothing
		table_ = Me
		sizeR_ = -1
		sizeC_ = -1
		sizing_ = 0

		allowResize_ = %IVResize.None
		copyHeaders_ = -1
		extensions_ = NewExtensionList()
		hLayout = NewSequentialLayout()
		vLayout = NewSequentialLayout()

		hNorm = LoadCursor( %NULL, ByVal %IDC_ARROW )
		hHorz = LoadCursor( GetModuleHandle( ByVal 0 ), "CUR_HORZ" )
		If IsFalse hHorz Then hHorz = LoadCursor( %NULL, ByVal %IDC_SIZEWE )
		hVert = LoadCursor( GetModuleHandle( ByVal 0 ), "CUR_VERT" )
		If IsFalse hVert Then hVert = LoadCursor( %NULL, ByVal %IDC_SIZENS )
		hBoth = LoadCursor( GetModuleHandle( ByVal 0 ), "CUR_BOTH" )
		If IsFalse hBoth Then hBoth = LoadCursor( %NULL, ByVal %IDC_SIZENWSE )

		drawControlBack_ = NewSysRenderFill( %COLOR_APPWORKSPACE )
		drawContentBack_ = no

		drawVGrid_ = no
		drawHGrid_ = no
	End Method

	Class Method FlagForPaint()
		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method
		InvalidateRect( hWnd_, ByVal %NULL, 0 )
	End Method

	Class Method GetContentRect( rc As RECT )
		Local lrc As RECT
		hLayout.GetRect( 0, lrc ):                      rc.Left = lrc.Left
		hLayout.GetRect( hLayout.VisibleCount-1, lrc ): rc.Right = lrc.Right
		vLayout.GetRect( 0, lrc ):                      rc.Top = rc.Top
		vLayout.GetRect( vLayout.VisibleCount-1, lrc ): rc.Bottom = lrc.Bottom
	End Method

	Class Method OnPaint()
		Local r As Long
		Local c As Long
		Local rc As RECT
		Local crc As RECT
		Local hdc As Dword
		Local mhdc As Dword
		Local ps As PAINTSTRUCT
		Local hImg As Dword
		Local pimg As Dword
		Local contentRc As RECT
		Local rowCount As Long
		Local colCount As Long
		Local ro As Long
		Local co As Long
		Local rn As iRender
		Local curCol As Long
		Local curRow As Long
		Local selCol As Long
		Local selRow As Long


		If IsFalse IsObject( table_ ) Then Exit Method
		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method
		hdc = BeginPaint( hWnd_, ps )
		hLayout.SuspendPaint = 1
		vLayout.SuspendPaint = 1
		GetClientRect( hWnd_, crc )
		hImg = CreateCompatibleBitmap( hdc, crc.Right, crc.Bottom )
		mhdc = CreateCompatibleDC( hdc )
		pimg = SelectObject( mhdc, hImg )

		CopyRect( rc, crc )
		vLayout.EnsureLayoutUpdated( rc )
		CopyRect( rc, crc )
		hLayout.EnsureLayoutUpdated( rc )

		rowCount = vLayout.VisibleCount-1
		colCount = hLayout.VisibleCount-1
		Me.GetContentRect( contentRc )

		If IsObject( drawControlBack_ ) Then
			CopyRect( rc, crc )
			drawControlBack_.Render( mhdc, rc, 0, 0 )
		End If
		If IsObject( drawContentBack_ ) Then
			CopyRect( rc, contentRc )
			drawContentBack_.Render( mhdc, rc, 0, 0 )
		End If

		' Render row backs
		vLayout.SetOtherBounds( contentRc.Left, contentRc.Right )
		vLayout.OtherItem = 0
		curRow = vLayout.CurItem
		selRow = vLayout.SelStart
		rn = vLayout
		If IsInterface( rn, iRender ) Then
			CopyRect( rc, crc )
			rn.Render( mhdc, rc, 0, 0 )
		End If

		curCol = hLayout.CurItem
		selCol = hLayout.SelStart
		hLayout.OtherLayout = vLayout

		' Draw cell contents
		rn = hLayout
		If IsInterface( rn, iRender ) Then
			hLayout.OtherItemCurrent = vLayout.CurItem
			For ro = 0 To rowCount
				vLayout.GetRect( ro, rc )
				r = vLayout.ItemIndex( ro )
				If Between( r,  selRow, curRow ) Or r=-1 Then
					hLayout.CurItem = curCol
					hLayout.SelStart = selCol
				Else
					hLayout.CurItem = -2
					hLayout.SelStart = -2
				End If
				hLayout.OtherItem = r
				hLayout.SetOtherBounds ( rc.Top, rc.bottom )
				rn.Render( mhdc, rc, r, 0 )

				If vLayout.XGutter > 0  And IsObject( drawHGrid_ ) Then
					' Draw horizontal gutter
					vLayout.GetRect( ro, rc )
					rc.Left = contentRc.Left
					rc.Right = contentRc.Right
					rc.Top = rc.Bottom
					rc.Bottom += vLayout.YGutter
					drawHGrid_.Render( mhdc, rc, r, 0 )
				End If
			Next
		End If
		hLayout.CurItem = curCol
		hLayout.SelStart = selCol
		' Draw vertical gutter
		If hLayout.YGutter>0 And IsInterface( drawVGrid_, iRender ) Then
			For co = 0 To colCount
				c = hLayout.ItemIndex( co )
				hLayout.GetRect( co, rc )
				rc.Top = contentRc.Top
				rc.Bottom = contentRc.Bottom
				rc.Left = rc.Right
				rc.Right += hLayout.XGutter
				drawVGrid_.Render( mhdc, rc, 0, c )
			Next
		End If
		hLayout.SetOtherBounds( crc.Top, crc.Bottom )

		Call BitBlt( hdc, 0, 0, crc.Right, crc.Bottom, mhdc, 0, 0, %SRCCOPY )
		SelectObject( mhdc, pimg )
		DeleteDc( mhdc )
		DeleteObject( hImg )
		EndPaint( hWnd_, ps )
		hLayout.SuspendPaint = 0
		vLayout.SuspendPaint = 0
	End Method

	Class Method OnTimer( ByVal timerId As Dword )
		Local cursor As PointAPI
		' Dragons here.  Just a note, but this is a low priority message
		' if you have message floods, this won't fire.  For example if
		' paint were to invalidate the window, this will never fire.


		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		If timerId = %IVDragTimer Then
			GetCursorPos( cursor )
			MapWindowPoints( %NULL, hWnd_, VarPtr( cursor ), 1 )
			SendMessage( hWnd_, %WM_MouseMove, 0, Mak( Dword, cursor.x, cursor.y )  )
		End If
	End Method

	Class Method OnCopy()
		Local r As Long
		Local c As Long
		Local sr As Long
		Local er As Long
		Local sc As Long
		Local ec As Long
		Local rn As iRender
		Local renderCell As Long
		Local renderRow As Long
		Local tx As WString
		Local sb As IStringBuilderW
		Local seq As iSequentialLayout


		If Not IsObject( base_ ) Then Exit Method
		If Not IsObject( dataSource_ ) Then Exit Method

		sb = Class "StringBuilderW"

		sr = vLayout.SelStart
		er = base_.Row
		If sr>er Then Swap sr, er

		hLayout.OtherLayout = vLayout
		seq = hLayout
		If IsObject(seq) And seq.IsRowSelection Then
			sc = 0
			ec = hLayout.ItemCount-1
		Else
			sc = hLayout.SelStart
			ec = base_.Col
		End If
		If sc>ec Then Swap sc, ec

		If copyHeaders_ Then
			For c = sc To ec
				If c<>sc Then sb.Add( $Tab )
				sb.add( dataSource_.Value( -1, c ) )
			Next c
			sb.Add( $CrLf )
		End If

		For r = sr To er
			For c = sc To ec
				If c<>sc Then sb.Add( $Tab )
				sb.Add( dataSource_.Value( r, c ) )
			Next
			sb.Add( $CrLf )
		Next
		Clipboard Reset
		Clipboard Set Text sb.String
	End Method

	Class Method OnActivate()
		Local rc As RECT

		If Not IsObject( base_ ) Then Exit Method

		Me.Goto( base_.Row, base_.Col )
		hLayout.SelStart = hLayout.CurItem
		vLayout.SeLStart = vLayout.CurItem

		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method
		GetClientRect( hWnd_, rc )
		Me.OnSize( Mak( Dword, rc.Right, rc.Bottom ) )
	End Method

	Class Method OnSize( ByVal wParam As Dword )
		Local si As SCROLLINFO
		Local rc As RECT

		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		GetClientRect( hWnd_, rc )
		hLayout.EnsureLayoutUpdated( rc )
		GetClientRect( hWnd_, rc )
		vLayout.EnsureLayoutUpdated( rc )

		si.cbSize = SizeOf( si )
		si.fMask = %SIF_All: '%SIF_Range Or %SIF_Page Or %SIF_Pos

		GetScrollInfo( hWnd_, %SB_Horz, si )
		si.nMin = hLayout.MinFirst
		si.nMax = hLayout.ItemCount - 1
		si.nPage = hLayout.PageItemCount
		si.nPos = hLayout.FirstItem
		SetScrollInfo( hWnd_, %SB_Horz, si, %TRUE )

		si.cbSize = SizeOf( si )
		si.fMask = %SIF_All: '%SIF_Range Or %SIF_Page Or %SIF_Pos
		GetScrollInfo( hWnd_, %SB_Vert, si )
		si.nMin = vLayout.MinFirst
		si.nMax = vLayout.ItemCount - 1
		si.nPage = vLayout.PageItemCount
		si.nPos = vLayout.FirstItem
		SetScrollInfo( hWnd_, %SB_Vert, si, %TRUE )
		Me.FlagForPaint()
	End Method

	Class Method OnHScroll( ByVal wParam As Long )
		hLayout.FirstItem = Me.OnScroll( %SB_Horz, wParam )
		Me.OnSize( wParam )
		Me.FlagForPaint()
	End Method

	Class Method OnVScroll( ByVal wParam As Long ) As String
		vLayout.FirstItem = Me.OnScroll( %SB_Vert, wParam )
		Me.OnSize( wParam )
		Me.FlagForPaint()
	End Method

	Class Method OnScroll( ByVal pDirection As Long, ByVal scrollType As Long ) As Long
		Local si As SCROLLINFO
		Local rc As RECT
		Local i As Long
		Local w As Long

		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		si.cbSize = SizeOf( si )
		si.fMask = %SIF_All
		GetScrollInfo( hWnd_, pDirection, si )
		Select Case LoWrd( scrollType )
			Case %SB_Top:        si.nPos = si.nMin
			Case %SB_Bottom:     si.nPos = si.nMax
			Case %SB_LineUp:     Decr si.nPos
			Case %SB_LineDown:   Incr si.nPos
			Case %SB_PageUp:     si.nPos -= si.nPage
			Case %SB_PageDown:   si.nPos += si.nPage
			Case %SB_ThumbTrack: si.nPos = si.nTrackPos
		End Select
		SetScrollPos( hWnd_, pDirection, si.nPos, %TRUE )
		Method = GetScrollPos( hWnd_, pDirection )
	End Method

	Class Method OnMouseWheel( ByVal wParam As Dword )
		Local dir As Dword

		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		If GetFocus()<>hWnd_ Then Exit Method
		dir = IIf&( Hi( Integer, wParam ) < 0, %SB_PageDown, %SB_PageUp )

		If ( Lo( Word, wParam ) And %MK_CONTROL )<>0 Then
			Call Me.OnHScroll( dir )
		Else
			Call Me.OnVScroll( dir )
		End If
	End Method

	Class Method OnMouseUp( ByVal x As Long, ByVal y As Long ) As String
		Local c As Long
		Local r As Long
		Local rc As RECT

		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		If dragging_ Then
			dragging_ = 0
			KillTimer( hWnd_, %IVDragTimer )
			ReleaseCapture()
			Exit Method
		End If

		If sizing_ Then
			sizing_ = 0
			ReleaseCapture()
			If sizeC_>= 0 Then
				c = hLayout.ItemIndex( sizeC_ )
				hLayout.GetRect( sizeC_, rc )
				hLayout.SizeOverride( c ) = x - rc.Left + sizeX_
			End If
			If sizeR_>= 0 Then
				r = vLayout.ItemIndex( sizeR_ )
				vLayout.GetRect( sizeR_, rc )
				vLayout.SizeOverride( r ) = y - rc.Top + sizeY_
			End If
			Me.SetCursorType( x, y )
			Exit Method
		End If
	End Method

	Class Method OnMouseDown( ByVal wParam As Long, ByVal x As Long, ByVal y As Long )
		Local r As Long
		Local c As Long
		Local rc As RECT
		Local sk As ShiftKeys

		If sizing_ Or dragging_ Then Exit Method

		c = hLayout.GetBorder( x, y )
		r = vLayout.GetBorder( x, y )
		If ( allowResize_ And %IVResize.Width ) = 0 Then c = -1
		If ( allowResize_ And %IVResize.Height ) = 0 Then r = -1
		If ( allowResize_ And %IVResize.HeaderOnly ) <> 0 Then
			If y<0 Or y>vLayout.SizeItem(0) Or vLayout.ItemIndex( 0 )<>-1 Then c=-1
			If x<0 Or x>hLayout.SizeItem(0) Or hLayout.ItemIndex( 0 )<>-1 Then r=-1
		End If

		If c>-1 Or r>-1 Then
			If c>-1 Then
				hLayout.GetRect( c, rc )
				sizeX_ = rc.Right - x: ' Distance from Right
				sizeC_ = c
			Else
				sizeC_ = -2
			End If
			If r>-1 Then
				vLayout.GetRect( r, rc )
				sizeY_ = rc.Bottom - y : ' Distance from Bottom
				sizeR_ = r
			Else
				sizeR_ = -2
			End If
			SetCapture( hWnd_ )
			sizing_ = -1
			Exit Method
		End If

		' Goto Click
		GetClientRect( hWnd_, rc )
		c = hLayout.Hit( x, y )
		c = hLayout.ItemIndex( c )
		If c<0 Then c = hLayout.CurItem

		' Span all content cols
		r = vLayout.Hit( x, y )
		r = vLayout.ItemIndex( r )
		If r<0 Then r = vLayout.CurItem


		Me.Goto( r, c )
		If ( wParam And %MK_SHIFT ) = 0 Then hLayout.selStart = c
		If ( wParam And %MK_SHIFT ) = 0 Then vLayout.SelStart = r
		dragging_ = -1
		SetCapture( hWnd_ )
		Me.FlagForPaint()
	End Method

	Class Method OnMouseMove( ByVal wParam As Long, ByVal x As Long, ByVal y As Long ) As String
		Local c As Long
		Local r As Long
		Local t As Long
		Local rc As RECT
		Local crc As RECT
		Local pt As Point
		Local adr As Dword

		If IsFalse IsObject( table_ ) Then Exit Method
		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		If hLayout.IsHotTracking Or vLayout.IsHotTracking Then Me.FlagForPaint()

		If dragging_ Then
			GetClientRect( hWnd_, crc )

			' Get past headers and frozen lines
			For t = 0 To hLayout.VisibleCount-1
				If hLayout.ItemIndex( t ) >= hLayout.FirstItem Then
					' Only scrolling if we're not up to first item
					If t>0 And hLayout.FirstItem > hLayout.ItemIndex( t-1 )+1 Then
						hLayout.GetRect( t, rc )
						crc.Left = rc.Left
					End If
					Exit For
				End If
			Next
			For t = 0 To vLayout.VisibleCount-1
				If vLayout.ItemIndex( t ) >= vLayout.FirstItem Then
					' Only scrolling if we're not up to first item
					If t>0 And vLayout.FirstItem > vLayout.ItemIndex( t-1 )+1 Then
						vLayout.GetRect( t, rc )
						crc.Top = rc.Top
					End If
					Exit For
				End If
			Next

			pt.x = x
			pt.y = y

			If PtInRect( crc, pt ) Then
				' Point is in rect
				' Span all content rows
				hLayout.SetOtherBounds( crc.Top, crc.Bottom )
				c = hLayout.Hit( x, y )
				If c<0 Then c = hLayout.VisibleCount-1
				c = hLayout.ItemIndex( c )
				If c<0 Then c = hLayout.CurItem

				vLayout.SetOtherBounds( crc.Left, crc.Right )
				r = vLayout.Hit( x, y )
				If r<0 Then r = vLayout.VisibleCount-1
				r = vLayout.ItemIndex( r )
				If r<0 Then r = vLayout.CurItem

				Me.Goto( r, c )
				Exit Method:
			End If

			' Only use timer if mouse isn't moving
			r = vLayout.CurItem
			c = hLayout.CurItem
			' Point is outside window
			If x<crc.Left Then
				Decr c
			ElseIf x>crc.Right Then
				Incr c
			End If
			If y<crc.Top Then
				Decr r
			ElseIf y>crc.Bottom Then
				Incr r
			End If
			Me.Goto( r, c )
			SetTimer( hWnd_, %IVDragTimer, 50, GetIV_ControlProc() )
			Exit Method
		End If

		If sizing_ Then
			GetClientRect( hWnd_, crc )
			If sizeC_>= 0 Then
				c = hLayout.ItemIndex( sizeC_ )
				hLayout.GetRect( sizeC_, rc )
				hLayout.SizeOverride( c ) = Max&( 1, x - rc.Left + sizeX_ )
				hLayout.RequestLayoutUpdate( crc )
			End If
			If sizeR_>= 0 Then
				r = vLayout.ItemIndex( sizeR_ )
				vLayout.GetRect( sizeR_, rc )
				vLayout.SizeOverride( r ) = Max&( 1, y - rc.Top + sizeY_ )
				vLayout.RequestLayoutUpdate( crc )
			End If
			Exit Method
		End If
		Call Me.SetCursorType( x, y )
	End Method

	Class Method SetCursorType( ByVal x As Long, ByVal y As Long )
		Local r As Long
		Local c As Long

		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		c = hLayout.GetBorder( x, y ): ' Offset
		r = vLayout.GetBorder( x, y ): ' Offset
		If ( allowResize_ And %IVResize.Width ) = 0 Then c = -1
		If ( allowResize_ And %IVResize.Height ) = 0 Then r = -1
		If ( allowResize_ And %IVResize.HeaderOnly ) <> 0 Then
			If y<0 Or y>vLayout.SizeItem(0) Or vLayout.ItemIndex( 0 )<>-1 Then c=-1
			If x<0 Or x>hLayout.SizeItem(0) Or hLayout.ItemIndex( 0 )<>-1 Then r=-1
		End If
		If     c< 0 And r <0 Then
			SetClassLong( hWnd_, %GCL_HCURSOR, hNorm )
		ElseIf c< 0 And r>= 0 Then
			SetClassLong( hWnd_, %GCL_HCURSOR, hVert )
		ElseIf c>= 0 And r< 0 Then
			SetClassLong( hWnd_, %GCL_HCURSOR, hHorz )
		ElseIf c>= 0 And r>= 0 Then
			SetClassLong( hWnd_, %GCL_HCURSOR, hBoth )
		End If
	End Method

	Class Method OnKeyPress( ByVal vKey As Dword ) As Long
		If vKey = 1 Or vKey = 27 Then
			Method = 0: ' 0 if handled
		Else
			Method = 1
		End If
	End Method

	Class Method OnKeyUp( ByVal vKey As Dword ) As Long
		Select Case vKey
			Case %VK_A, %VK_C, %VK_ESCAPE
				Method = 0
			Case Else
				Method = 1
		End Select
	End Method

	Class Method OnKeyDown( ByVal vKey As Dword ) As Long
		Local si As SCROLLINFO
		Local sk As ShiftKeys
		Local i As Long
		Local v As Long
		Local rc As RECT
		Local r As Long
		Local c As Long

		If IsFalse IsObject( base_ ) Then Exit Method
		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		Call GetShifts( sk )
		r = base_.Row
		c = base_.Col

		Method = 1: ' Unhandled
		Select Case vKey
			Case %VK_ESCAPE
					vLayout.SelStart = vLayout.CurItem
					hLayout.SelStart = hLayout.CurItem
					Me.FlagForPaint()
					Exit Method
			Case %VK_A
				If sk.isCtrl And Not ( sk.isShift Or sk.isAlt ) Then
					GetClientRect( hWnd_, rc )
					vLayout.SelStart = vLayout.ItemCount-1
					vLayout.CurItem = 0
					vLayout.FirstItem = 0
					base_.Row = 0

					hLayout.SelStart = hLayout.ItemCount-1
					hLayout.CurItem = 0
					hLayout.FirstItem = 0
					base_.Col = 0
					Exit Method
				End If
			Case %VK_C
				If sk.isCtrl And Not ( sk.isShift Or sk.isAlt ) Then
					Call Me.OnCopy()
					Exit Method
				End If
			Case %VK_LEFT:
				If sk.isCtrl And Not ( sk.isAlt ) Then
					si.cbSize = SizeOf( si )
					si.fMask  = %SIF_All
					GetScrollInfo( hWnd_, %SB_Horz, si )
					c -= ( si.nPage )
				ElseIf Not ( sk.isCtrl Or sk.isAlt ) Then
					Decr c
				Else
					Exit Method
				End If
			Case %VK_RIGHT:
				If sk.isCtrl And Not ( sk.isAlt ) Then
					si.cbSize = SizeOf( si )
					si.fMask  = %SIF_All
					GetScrollInfo( hWnd_, %SB_Horz, si )
					c += ( si.nPage )
				ElseIf Not ( sk.isCtrl Or sk.isAlt ) Then
					Incr c
				Else
					Exit Method
				End If
			Case %VK_UP:
				If Not ( sk.isCtrl Or sk.isAlt ) Then
					Decr r
				ElseIf sk.isCtrl And Not sk.isAlt Then
					r = 0
				Else
					Exit Method
				End If
			Case %VK_DOWN:
				If Not ( sk.isCtrl Or sk.isAlt ) Then
					Incr r
				ElseIf sk.isCtrl And Not sk.isAlt Then
					r = base_.RowCount - 1
				Else
					Exit Method
				End If
			Case %VK_END:
				If Not ( sk.isCtrl Or sk.isAlt ) Then
					c = base_.ColCount - 1
				ElseIf sk.isCtrl And Not ( sk.isAlt ) Then
					c = base_.ColCount - 1
					r = base_.RowCount - 1
				Else
					Exit Method
				End If
			Case %VK_HOME
				If Not ( sk.isCtrl Or sk.isAlt ) Then
					c = 0
				ElseIf sk.isCtrl And Not ( sk.isAlt ) And ( r Or c ) Then
					c = 0
					r = 0
				Else
					Exit Method
				End If
			Case %VK_PRIOR
				si.cbSize = SizeOf( si )
				si.fMask  = %SIF_All
				GetScrollInfo( hWnd_, %SB_Vert, si )
				r -= Max&( si.nPage-2, 10 )
			Case %VK_NEXT
				si.cbSize = SizeOf( si )
				si.fMask  = %SIF_All
				GetScrollInfo( hWnd_, %SB_Vert, si )
				r += Max&( si.nPage-2, 10 )
			Case Else
				Exit Method
		End Select
		Me.Goto( r, c )
		If IsFalse sk.isShift Then
			vLayout.SelStart = vLayout.CurItem
			hLayout.SelStart = hLayout.CurItem
		End If
		Method = 0
	End Method

	Class Method GoTo( ByVal r As Long, ByVal c As Long )
		Local msg As TblNotify
		Local rc As RECT
		Local lr As Long
		Local lc As Long

		If IsFalse IsObject( table_ ) Then Exit Method
		If IsFalse IsObject( base_ ) Then Exit Method
		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method

		r = Max&( 0, Min&( vLayout.ItemCount-1, r ) )
		c = Max&( 0, Min&( hLayout.ItemCount-1, c ) )

		lc = hLayout.CurItem
		If lc<>c And c>-1 Then
			hLayout.GotoItem( c )
			SetScrollPos( hWnd_, %SB_Horz, hLayout.FirstItem, %TRUE )
			base_.Col = c
		End If

		lr = vLayout.CurItem
		If lr<>r And r>-1 Then
			vLayout.GotoItem( r )
			SetScrollPos( hWnd_, %SB_Vert, vLayout.FirstItem, %TRUE )
			base_.Row = r
		End If

		If lr<>r Or lc<>c Then
			Me.FlagForPaint()
			msg.hdr.code = %TBN_CellChg
			msg.hdr.idFrom = GetDlgCtrlID( hWnd_ )
			msg.hdr.hwndFrom = hWnd_
			msg.r = r
			msg.c = c

			SendMessage( GetParent( hWnd_ ), %WM_Notify, hWnd_, VarPtr( msg )  )
		End If
	End Method

	Interface iIVRowCol Guid$( "{770c3baf-c13b-497f-ac3e-beec83fe815d}" )
		Inherit IUnknown

		Property Get AllowResize() As Long
			Property = allowResize_
		End Property
		Property Set AllowResize( ByVal value As Long )
			Local cursor As Point
			value And= 7
			If value<> allowResize_ Then
				allowResize_ = value
				GetCursorPos( cursor )
				MapWindowPoints( %NULL, hWnd_, VarPtr( cursor ), 1 )
				Me.SetCursorType( cursor.X, cursor.Y )
			End If
		End Property

		Property Get HLayout() As iLayout
			Property = hLayout
		End Property
		Property Set HLayout( ByRef value As iLayout )
			Local rc As RECT
			If IsInterface( value, iLayout ) And ObjPtr( value ) <> ObjPtr( hLayout ) Then
				hLayout = value
				hLayout.HWnd = hWnd_
				If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Property
				GetClientRect( hWnd_, rc )
				hLayout.RequestLayoutUpdate( rc )
			End If
		End Property

		Property Get VLayout() As iLayout
			Property = vLayout
		End Property
		Property Set VLayout( ByRef value As iLayout )
			Local rc As RECT
			If IsInterface( value, iLayout ) And ObjPtr( value ) <> ObjPtr( vLayout ) Then
				vLayout = value
				vLayout.HWnd = hWnd_
				If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Property
				GetClientRect( hWnd_, rc )
				vLayout.RequestLayoutUpdate( rc )
			End If
		End Property

		Property Get CopyHeaders() As Long
			Property = copyHeaders_
		End Property
		Property Set CopyHeaders( ByVal value As Long )
			copyHeaders_ = IIf&( value, -1, 0 )
		End Property

		Property Get DrawVGrid() As iRender
			Property = drawVGrid_
		End Property
		Property Set DrawVGrid( ByRef value As iRender )
			If ObjPtr( value ) <> ObjPtr( drawVGrid_ ) Then
				drawVGrid_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get DrawHGrid() As iRender
			Property = drawHGrid_
		End Property
		Property Set DrawHGrid( ByRef value As iRender )
			If ObjPtr( value ) <> ObjPtr( drawHGrid_ ) Then
				drawHGrid_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get DrawContentBack() As iRender
			Property = drawContentBack_
		End Property
		Property Set DrawContentBack( ByRef value As iRender )
			If ObjPtr( value ) <> ObjPtr( drawContentBack_ ) Then
				drawContentBack_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get DrawControlBack() As iRender
			Property = drawControlBack_
		End Property
		Property Set DrawControlBack( ByRef value As iRender )
			If ObjPtr( value ) <> ObjPtr( drawControlBack_ ) Then
				drawControlBack_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get Extensions() As iExtensionList
			Property = extensions_
		End Property
	End Interface

	Interface iExtend Guid$( "{071a4670-b6c6-4fd2-8d44-5d4caa0af66b}" )
		Inherit IUnknown

		Method Extend( ByRef value As iWin )
			Local hds As iHasDataSource
			Local meHds As iHasDataSource
			Local ds As iDataSource


			base_ = value
			hds = value
			meHds = Me
			If IsTrue IsInterface( value, iWin ) Then
				hWnd_ = value.HWnd
				hLayout.HWnd = hWnd_
				vLayout.HWnd = hWnd_
			End If
			If IsInterface( hds, iHasDataSource ) And IsInterface( meHds, iHasDataSource ) Then
				ds = hds.Datasource
				meHds.Datasource = ds
			End If
		End Method

		Method Unextend()
			base_ = Nothing
			table_ = Nothing
			dataSource_ = Nothing
		End Method

		Method Execute( ByVal wMsg As Dword, ByVal wParam As Long, ByVal lParam As Long, ByRef ret As Long ) As Long
			Method = -1: ' Handled
			' ret already defaulting to 0

			' Check if extension is handling message
			If IsTrue IsObject( extensions_ ) Then
				extensions_.Reset()
				While extensions_.Next
					If extensions_.Current.Execute( wMsg, wParam, lParam, ret ) Then Exit Method
				Wend
			End If

			Select Case As Const wMsg
				Case %WM_Activate:        Call Me.OnActivate(): Exit Method
				Case %WM_Paint:           Call Me.OnPaint(): Exit Method
				Case %WM_Timer:           Call Me.OnTimer( wParam ): Exit Method
				Case %WM_Size:            Call Me.OnSize( lParam ): Exit Method
				Case %WM_HScroll:         Call Me.OnHScroll( wParam ): Exit Method
				Case %WM_VScroll:         Call Me.OnVScroll( wParam ): Exit Method
				Case %WM_MouseWheel:      Call Me.OnMouseWheel( wParam ): Exit Method
				Case %WM_KeyDown:         ret = Me.OnKeyDown( wParam ): Exit Method
				Case %WM_Char:            ret = Me.OnKeyPress( wParam ): Exit Method
				Case %WM_KeyUp:           ret = Me.OnKeyUp( wParam ): Exit Method
				Case %WM_LButtonDown:     Call Me.OnMouseDown( wParam, Lo( Integer, lParam ), Hi( Integer, lParam ) ): Exit Method
				Case %WM_LButtonUp:       Call Me.OnMouseUp( Lo( Integer, lParam ), Hi( Integer, lParam ) ): Exit Method
				Case %WM_MouseMove:       Call Me.OnMouseMove( wParam, Lo( Integer, lParam ), Hi( Integer, lParam ) ): Exit Method
			End Select
			Method = 0: ' Unhandled
		End Method
	End Interface

	Interface iHasDatasource Guid$( "{33e74a9a-2739-4180-a894-3fc9cd8614b3}" )
		Inherit IUnknown

		Property Get DataSource() As iDataSource
			Property = dataSource_
		End Property
		Property Set DataSource( ByRef value As iDataSource )
			Local hds As iHasDatasource
			dataSource_ = value
			hds = vLayout: If IsTrue IsInterface( hds, iHasDataSource ) Then hds.DataSource = value
			hds = hLayout: If IsTrue IsInterface( hds, iHasDataSource ) Then hds.DataSource = value
		End Property
	End Interface
End Class

Function Between( ByVal value As Long, ByVal a As Long, ByVal b As Long ) Common As Long
	If a > b Then Swap a, b
	Function = IIf&( value >= a And value <= b, -1, 0 )
End Function
