'#Include Once "datatable\DataTableShared.inc"
'#Include Once "datatable\DataColumn.inc"
'#Include Once "datatable\DataRow.inc"

Function NewDataTable( ByRef tableName As String, Optional ByRef ds As iDataSet ) Common Export As iDataTable
	Dim t As iDataTable
	t = Class "cDataTable"
	t.TableName = tableName
	If IsFalse IsMissing( ds) And IsInterface( ds, iDataSet ) Then
		ds.Add( t )
	End If
	Function = t
End Function

Sub AddDataTableRow( ByRef table As iDataTable, values() As String ) Common Export
	Local i As Long
	Local cnt As Long
	Local r As iDataRow

	cnt = UBound( values() ) + 1
	If Not IsInterface( table, iDataTable ) Then Exit Sub
	If cnt > table.ColCount Then cnt = table.ColCount

	r = table.NewRow()

	For i=0 To cnt-1
		r.Value(i) = values(i)
	Next

	Call table.AddRow( r )
End Sub

Class cDataTable Guid$("{4c7fbc6e-52f9-4387-9c65-ebfd76c4eebe}") Common
	Instance row_() As iDataRow
	Instance rows_ As Long
	Instance column_() As iDataColumn
	Instance columns_ As Long
	Instance version_ As Long

	Instance capacity_ As Long
	Instance tableName_ As String
	Instance dataSet_ As iDataSet
	Instance nullDisplayValue_ As String
	Instance send_ As iSend
	Instance suspendNotice_ As Long

	Class Method Create()
		nullDisplayValue_ = ""
		send_ = NewSend()
	End Method


	Class Method Publish( ByVal msg As Long, ByVal a As Long, ByVal b As Long)
		Local s As iSend
		Local publish_ As iPublish
		s = Me
		publish_ = send_
		publish_.Publish( s, msg, a, b )
	End Method

	Interface iDataTable Guid$("{a35004f4-ffe7-403d-8975-3cdb203fa7fe}")
		Inherit IUnknown

		Property Get TableName() As String
			Property = tableName_
		End Property
		Property Set TableName( ByRef value As String )
			tableName_ = value
		End Property

		Property Get Capacity() As Long
			Property = capacity_
		End Property
		Property Set Capacity( ByVal pVal As Long )
			If pVal>0 Then
				ReDim Preserve row_( pVal-1 )
				capacity_ = pVal
				If capacity_ < rows_ Then rows_ = capacity_
			Else
				Erase row_()
				capacity_ = 0
				rows_ = 0
			End If
		End Property

		Property Get ColCount() As Long
			Property = columns_
		End Property

		Method AddDelimetedColumns( ByRef delimetedColumns As String, ByVal delimeter As String )
			Local i As Long
			Local cnt As Long
			Local t As iDataTable
			Local colNames() As String
			cnt = ParseCount( delimetedColumns, delimeter )
			ReDim colNames(cnt-1)
			Parse delimetedColumns, colNames(), delimeter
			t = Me
			suspendNotice_ = -1
			For i=0 To cnt-1
				Call t.AddColumn( colNames(i) )
			Next
			suspendNotice_ = 0
			Incr version_
			If cnt>0 Then Me.Publish( %Message.CountChanged, rows_, columns_ )
		End Method

		Method AddColumn( ByRef columnName As String, Optional ByRef heading As String ) As iDataColumn
			Dim c As iDataColumn
			Dim t As iDataTable
			Dim sn As Long

			sn = suspendNotice_

			suspendNotice_ = -1
			Me.ClearRows()
			suspendNotice_ = sn
			If columns_ >= UBound( column_() ) Then
				ReDim Preserve column_( columns_ + 10 )
			End If

			c = NewDataColumn()
			t = Me
			c.DataTable = t
			c.ColumnName = columnName
			If Not IsMissing( heading) Then
				c.Heading = heading
			End If
			column_( columns_ ) = c
			Incr columns_
			Incr version_
			If Not suspendNotice_ Then Me.Publish( %Message.CountChanged, rows_, columns_ )
		End Method

		Method RemoveColumn( ByVal offset As Long )
			Local i As Long
			If columns_<1 Then Exit Method
			Me.ClearRows()
			Decr columns_
			For i=offset To columns_-1
				column_(i) = column_(i+1)
			Next
			column_(columns_) = Nothing
			Incr version_
			If Not suspendNotice_ Then Me.Publish( %Message.CountChanged, rows_, columns_ )
		End Method

		Property Get Column(ByVal offset As Long ) As iDataColumn
			If offset>=0 And offset<columns_ Then
				Property = column_( offset )
			End If
		End Property

		Property Get ColumnIndex( ByVal pName As String ) As Long
			Local i As Long
			Local nm As String
			pName = UCase$( pName )
			For i=0 To columns_ - 1
				nm = column_(i).ColumnName
				If pName = UCase$( nm ) Then
					Property = i
					Exit Property
				End If
			Next
			Property = -1
		End Property
		Property Get ColumnName( ByVal offset As Long ) As String
			If offset>=0 And offset<columns_ Then
				Property = column_( offset ).ColumnName
			End If
		End Property

		Method Clear()
			Erase row_()
			Erase column_()
			rows_ = 0
			columns_ = 0
			capacity_ = 0
			Incr version_
			If Not suspendNotice_ Then
				Me.Publish( %Message.CountChanged, rows_, columns_ )
			End If
		End Method

		Method ClearRows()
			Local i As Long
			For i = 0 To rows_ - 1
				row_(i) = Nothing
			Next
			rows_ = 0
			Incr version_
			If Not suspendNotice_ Then Me.Publish( %Message.CountChanged, rows_, columns_ )
		End Method

		Method NewRow() As iDataRow
			' Creates a row bound to table, but not added to table
			If columns_ = 0 Then Exit Method
			Dim r As iDataRow
			Dim t As iDataTable
			r = NewDataRow()
			t = Me
			r.DataTable = t
			Method = r
		End Method

		Method AddRow( ByRef value As iDataRow )
			Dim t As iDataTable
			If Not IsInterface( value, iDataRow ) Then Exit Method
			Dim incommingDataTable As iDataTable
			t = Me
			incommingDataTable = value.DataTable
			If ObjPtr( incommingDataTable ) <> ObjPtr( t ) Then Exit Method: ' Not bound to this DataTable
			If rows_>= capacity_ Then Me.Capacity = capacity_ + 1000
			row_(rows_) = value
			Incr rows_
			Incr version_
			If Not suspendNotice_ Then Me.Publish( %Message.CountChanged, rows_, columns_ )
		End Method

		Method RemoveRow( ByVal offset As Long )
			Local i As Long
			If offset>=0 And offset<rows_ Then
				Decr rows_
				For i=offset To rows_-1
					row_(i) = row_(i+1)
				Next
				row_(rows_) = Nothing
				Incr version_
				If Not suspendNotice_ Then Me.Publish( %Message.CountChanged, rows_, columns_ )
			End If
		End Method

		Method AddNewDelimetedRow( ByRef delimetedValues As String, ByVal delimeter As String )
			Local i As Long
			Local cnt As Long
			Local value() As String
			Local r As iDataRow
			Dim t As iDataTable
			Dim sn As Long

			cnt = ParseCount( delimetedValues, delimeter )
			If cnt>columns_ Then cnt=columns_

			ReDim value(cnt-1)
			Parse delimetedValues, value(), delimeter
			t = Me

			r = t.NewRow()

			For i=0 To cnt-1
				r.Value(i) = value(i)
			Next
			sn = suspendNotice_
			suspendNotice_ = -1
			Call t.AddRow( r )
			suspendNotice_ = sn
			If Not suspendNotice_ Then Me.Publish( %Message.CountChanged, rows_, columns_ )
		End Method

		Property Get RowCount() As Long
			Property = rows_
		End Property

		Property Get Row( ByVal r As Long ) As iDataRow
			If r>=0 And r<rows_ Then
				Property = row_( r )
			End If
		End Property
		Property Get NullDisplayValue() As String
			Property = nullDisplayValue_
		End Property
		Property Set NullDisplayValue( ByRef value As String )
			nullDisplayValue_ = value
		End Property
		Property Get DataSet() As iDataSet
			Property = dataSet_
		End Property
		Property Set DataSet( ByRef value As iDataSet )
			dataSet_ = value
		End Property

		Property Get Version() As Long
			Property = version_
		End Property
		Method NextVersion()
			Incr version_
		End Method
	End Interface

	Interface iSend Guid$("{7b1b3f52-c3bc-40b0-bf3f-9e239ed67dc8}")
		Inherit IUnknown
		Method Subscribe( value As iReceive )
			send_.Subscribe( value )
		End Method
		Method Unsubscribe( value As iReceive )
			send_.Unsubscribe( value )
		End Method
	End Interface
End Class
