Function NewFlowLayout() Common Export As iLayout
	Local o As iLayout
	o = Class "cFlowLayout"
	Function = o
End Function

Class cFlowLayout Guid$("{206e41de-92f0-4fab-9042-c94a45e167d0}") Common
	Instance visibleCount_ As Long:                   ' # of visible items
	Instance curItem_ As Long:                        ' current Item ( Current column or row #)
	Instance otherItem_ As Long:                      ' other Item ( Current row or column #)
	Instance otherItemCurrent_ As Long:               ' other item current (for focus)
	Instance otherLayout_ As iLayout                  ' other item layout
	Instance selStart_ As Long:                       ' Selection start
	Instance firstItem_ As Long:                      ' First unfrozen item ( Column or Row #)
	Instance maxFirst_ As Long:                       ' Maximum index of firstItem
	Instance isCurRow_ As Long:                       ' 0 = CurItem is columns, <>0 = CurItem is rows
	Instance pageSize_ As Long:                       ' Size of a page in items
	Instance suspendPaint_ As Long:                   ' <>0 = suspend painting
	Instance padLeft_ As Long
	Instance padRight_ As Long
	Instance padTop_ As Long
	Instance padBottom_ As Long

	Instance dataSource_ As iDataSource:              ' Data source
	Instance selType_ As Long:                        ' Selection is None, Cell, or Row
	Instance hotType_ As Long:                        ' Hot tracking is None, Cell, or Row
	Instance hWnd_ As Dword:                          ' Window doing layout in

	Instance bounds_ As RECT:                         ' Bounds to render inside of
	Instance rc_() As RECT:                           ' Locations of each item
	Instance itemIndex_() As Long:                    ' Offsets of items ( Column or Row #'s)

	Instance xgutter_ As Long                         ' Gutter size ( Distance between items, i.e. Grid )
	Instance ygutter_ As Long                         ' Gutter size ( Distance between items, i.e. Grid )

	Instance width_ As Long:                          '  > 0 = Pixel, < 0 = # columns
	Instance height_ As Long:                         '  > 0 = Pixel, < 0 = # rows
	Instance isRightDown_ As Long:                    '  <>0 = Flow right/down, =0 Flow down/right

	Instance backDefault_ As iStateRender:            ' Item back renderers ( back renderer for non-overridden backs)
	Instance backOverride_ As iLongStateRenderMap:    ' Item back renderer overrides ( back renderer for overridden backs)

	Instance contentDefault_ As iStateRender:         ' Content renderers ( content renderer for non-overridden contents)
	Instance contentOverride_ As iLongStateRenderMap: ' Item content renderer overrides ( content renderer for overridden contents)

	Instance drawFocus_ As iRender:                   ' Draw focus rectangle
	Instance drawOverride_ As iUserRender:            ' User rendering override
	Instance needLayout_ As Long:                     ' <>0 if layout needed

	Class Method Create()
		Dim rc_( 100)
		Dim itemIndex_( 100)

		width_ = -3: ' 3 columns
		height_ = 200: ' 200 pixels
		isRightDown_ = -1: ' Flow right then down
		isCurRow_ = -1:    ' Flow rows

		backDefault_ = NewStateRender()
		backOverride_ = NewLongStateRenderMap()

		contentDefault_ = NewStateRender()
		contentOverride_ = NewLongStateRenderMap()
	End Method

	Class Method SetHWnds()
		Local key As Long
		Local i As Long
		Local srm As iStateRender

		If IsObject( otherLayout_ ) Then otherLayout_.HWnd = hWnd_

		For i=0 To backOverride_.Count-1
			key = backOverride_.Key(i)
			srm = backOverride_.Value(key)
			Call Me.SetStateRenderHWnd( srm )
		Next
		For i=0 To contentOverride_.Count-1
			key = contentOverride_.Key(i)
			srm = contentOverride_.Value(key)
			Call Me.SetStateRenderHWnd( srm )
		Next

		Call Me.SetStateRenderHWnd( backDefault_ )
		Call Me.SetStateRenderHWnd( contentDefault_ )
		Call Me.SetRenderHWnd( drawFocus_ )
	End Method

	Class Method SetStateRenderHWnd( srm As iStateRender )
		Local rn As iRender
		If IsInterface( srm, iStateRender ) Then
			rn = srm.Item( %IVState.Normal ):   Call Me.SetRenderHWnd( rn )
			rn = srm.Item( %IVState.Selected ): Call Me.SetRenderHWnd( rn )
			rn = srm.Item( %IVState.Hot ):      Call Me.SetRenderHWnd( rn )
		End If
	End Method
	Class Method SetRenderHWnd( rn As iRender )
		Local lay As iLayout

		lay = rn
		If IsInterface( lay, iLayout ) Then lay.HWnd = hWnd_
	End Method

	Class Method ExtendLayout( ByRef ub As Long )
		ub += 100
		ReDim Preserve rc_( ub )
		ReDim Preserve itemIndex_( ub )
	End Method

	Class Method PaintPart( ByVal hdc As Long, ByRef rc As RECT, ByVal r As Long, ByVal c As Long, ByRef value As WString, ByVal part As Long, ByVal state As Long, rn As iRender )
		Local userHandled As Long
		Local v As iValue
		Local lay As iLayout
		Local sn As Long

		If IsInterface( drawOverride_, iUserRender ) Then
			userHandled = drawOverride_.Render( hdc, rc, r, c, value, part, State, rn )
		End If
		If IsFalse userHandled And IsInterface( rn, iRender ) Then
			v = rn: If IsObject( v ) Then v.Value = value
			lay = rn
			If IsInterface( lay, iLayout ) Then
				If isCurRow_ Then
					lay.OtherItem = r
				Else
					lay.OtherItem = c
				End If
				sn = lay.SuspendPaint
				lay.SuspendPaint=1
				lay.EnsureLayoutUpdated( rc )
				lay.SetOtherBounds( rc.Left, rc.Right )
				lay.SuspendPaint = sn
			End If
			rn.Render( hdc, rc, r, c )
		End If
	End Method

	Class Method FlagForPaint()
		If suspendPaint_ Or IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method
		InvalidateRect( hWnd_, bounds_, 0 )
	End Method

	Class Method FlagForLayout()
		needLayout_ = -1
		Me.FlagForPaint()
	End Method

	Class Method DoUpdateLayout()
		Local i As Long
		Local x As Long
		Local y As Long
		Local cardWidth As Long
		Local cardHeight As Long
		Local ub As Long
		Local seq As iFlowLayout
		Local lay As iLayout
		Local cnt As Long
		Local b As Long

		needLayout_ = 0
		seq = Me
		lay = Me
		visibleCount_ = 0
		x = bounds_.Left + padLeft_
		y = bounds_.Top + padTop_
		cardHeight = height_
		cardWidth = width_
		If cardHeight<0 Then
			cardHeight = -( (bounds_.Bottom-bounds_.Top - padTop_ - padBottom_ + ygutter_) \ cardHeight ) - ygutter_
		End If
		If cardWidth<0 Then
			cardWidth = -( (bounds_.Right-bounds_.Left - padLeft_ - padRight_ + xgutter_) \ cardWidth ) - xgutter_
		End If
		Local rows As Long
		Local cols As Long
		If cardHeight+ygutter_ Then
			rows = (bounds_.Bottom-bounds_.Top - padTop_ - padBottom_ + ygutter_ ) \ ( cardHeight + ygutter_ )
		Else
			rows = 1
		End If
		If cardWidth+xgutter_ Then
			cols = (bounds_.Right-bounds_.Left - padLeft_ - padRight_ + xgutter_ ) \ ( cardWidth + xgutter_ )
		Else
			cols = 1
		End If
		If cols=1 Then
			cols=1
		End If
		ub = UBound( rc_() )
		visibleCount_ = 0
		i = firstItem_
		cnt = lay.ItemCount
		If isRightDown_ Then
			While y + cardHeight  <= bounds_.Bottom - padBottom_ And i < cnt
				While x + cardWidth <= bounds_.Right - padRight_ And i < cnt
					If visibleCount_ > ub Then Me.ExtendLayout( ub )
					itemIndex_( visibleCount_ ) = i
					SetRect( rc_( visibleCount_ ), x, y, x+cardWidth, y+cardHeight )
					Incr i
					Incr visibleCount_
					x += cardWidth + xgutter_
				Wend
				x = bounds_.Left + padLeft_
				y += cardHeight + ygutter_
			Wend
		Else
			While x + cardWidth <= bounds_.Right - padRight_ And i < cnt
				While y + cardHeight <= bounds_.Bottom - padBottom_ And i < cnt
					If visibleCount_ > ub Then Me.ExtendLayout( ub )
					itemIndex_( visibleCount_ ) = i
					SetRect( rc_( visibleCount_ ), x, y, x+cardWidth, y+cardHeight )
					Incr i
					Incr visibleCount_
					y += cardHeight + ygutter_
				Wend
				y = bounds_.Top + padTop_
				x += cardWidth + xgutter_
			Wend
		End If
		pageSize_ = rows * cols
		maxFirst_ = Max&(0, lay.ItemCount - pageSize_)
		Me.FlagForPaint()
	End Method

	Interface iFlowLayout Guid$("{373554bb-cbd0-4a55-a555-c1edbb274b1d}")
		Inherit IUnknown

		Property Get IsRightDown() As Long
			Property = isRightDown_
		End Property
		Property Set IsRightDown( ByVal value As Long )
			isRightDown_ = IIf&( value, -1, 0 )
			Me.FlagForLayout()
		End Property

		Property Get IsCurItemARow() As Long
			Property = isCurRow_
		End Property
		Property Set IsCurItemARow( ByVal value As Long )
			isCurRow_ = IIf&( value, -1, 0 )
			Me.FlagForLayout()
		End Property

		Property Get SelectType() As Long
			Property = selType_
		End Property
		Property Set SelectType( ByVal value As Long )
			Select Case value
				Case %IVSelect.None, %IVSelect.Content
					selType_ = value
			End Select
		End Property

		Property Get HotType() As Long
			Property = hotType_
		End Property
		Property Set HotType( ByVal value As Long )
			Select Case value
				Case %IVSelect.None, %IVSelect.Content
					hotType_ = value
			End Select
		End Property

		Property Get BackRenderer( ByVal offset As Long, ByVal state As Long ) As iRender
			Local r As iStateRender
			Local l As iLayout
			l = Me
			r = backOverride_.Value( offset )
			If Not IsObject( r ) Then r = backDefault_
			If IsObject( r ) Then Property = r.Item( State )
		End Property
		Property Get ContentRenderer( ByVal offset As Long, ByVal state As Long ) As iRender
			Local r As iStateRender
			Local l As iLayout
			l = Me
			r = contentOverride_.Value( offset )
			If Not IsObject( r ) Then r = contentDefault_
			If IsObject( r ) Then Property = r.Item( State )
		End Property

		Property Get DrawFocus() As iRender
			Property = drawFocus_
		End Property
		Property Set DrawFocus( ByRef value As iRender )
			If ObjPtr( value ) <> ObjPtr( drawFocus_ ) Then
				drawFocus_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get DrawOverride() As iUserRender
			Property = drawOverride_
		End Property
		Property Set DrawOverride( value As iUserRender )
			If ObjPtr( value ) <> ObjPtr( drawOverride_ ) Then
				drawOverride_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get BackDefault( ByVal state As Long ) As iRender
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				Property = backDefault_.Item( State )
			End If
		End Property
		Property Set BackDefault( ByVal state As Long, ByRef value As iRender )
			Local l As iLayout
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				backDefault_.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property

		Property Get BackOverride( ByVal offset As Long, ByVal state As Long ) As iRender
			Local rs As iStateRender
			Local l As iLayout
			l = Me
			If offset>=-1 And offset<l.ItemCount And State>=%IVState.Normal And State<=%IVState.Hot Then
				If backOverride_.Contains( offset ) Then
					rs = backOverride_.Value( offset )
					If IsObject( rs ) Then Property = rs.Item( State )
				End If
			End If
		End Property
		Property Set BackOverride( ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
			Local rs As iStateRender
			Local l As iLayout
			l = Me
			If offset>=-1 And offset<l.ItemCount And State>=%IVState.Normal And State<=%IVState.Hot Then
				If Not backOverride_.Contains( offset ) Then
					rs = NewStateRender()
					backOverride_.Add( offset, rs )
				Else
					rs = backOverride_.Value( offset )
				End If
				If IsObject( rs ) Then rs.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property
		Method BackOverrideClear( ByVal offset As Long )
			backOverride_.Remove( offset )
		End Method

		Property Get ContentDefault( ByVal state As Long ) As iRender
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				Property = contentDefault_.Item( State )
			End If
		End Property
		Property Set ContentDefault( ByVal state As Long, ByRef value As iRender )
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				contentDefault_.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property

		Property Get ContentOverride( ByVal offset As Long, ByVal state As Long ) As iRender
			Local rs As iStateRender
			Local l As iLayout
			l = Me
			If offset>=-1 And offset<l.ItemCount And State>=%IVState.Normal And State<=%IVState.Hot Then
				If contentOverride_.Contains( offset ) Then
					rs = contentOverride_.Value( offset )
					If IsObject( rs ) Then Property = rs.Item( State )
				End If
			End If
		End Property
		Property Set ContentOverride( ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
			Local rs As iStateRender
			Local l As iLayout
			l = Me
			If offset>=-1 And offset<l.ItemCount And State>=%IVState.Normal And State<=%IVState.Hot Then
				If Not contentOverride_.Contains( offset ) Then
					rs = NewStateRender()
					contentOverride_.Add( offset, rs )
				Else
					rs = contentOverride_.Value( offset )
				End If
				If IsObject( rs ) Then rs.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property
		Method ContentOverrideClear( ByVal offset As Long )
			If contentOverride_.Contains( offset ) Then
				contentOverride_.Remove( offset )
				Me.FlagForPaint()
			End If
		End Method

		Property Get Width() As Long
			Property = width_
		End Property
		Property Set Width( ByRef value As Long )
			If value<>0 And value<>width_ Then
				width_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get Height() As Long
			Property = height_
		End Property
		Property Set Height( ByRef value As Long )
			If value<>0 And value<>height_ Then
				height_ = value
				Me.FlagForLayout()
			End If
		End Property
	End Interface

	Interface iLayout Guid$( "{b87daa5f-0140-4b31-8bd1-ff4f0c56e666}")
		Inherit IUnknown

		Property Get PadLeft() As Long
			Property = padleft_
		End Property
		Property Set PadLeft( ByRef value As Long )
			padleft_ = value
		End Property

		Property Get PadTop() As Long
			Property = padtop_
		End Property
		Property Set PadTop( ByRef value As Long )
			padtop_ = value
		End Property

		Property Get PadRight() As Long
			Property = padright_
		End Property
		Property Set PadRight( ByRef value As Long )
			padright_ = value
		End Property

		Property Get PadBottom() As Long
			Property = padbottom_
		End Property
		Property Set PadBottom( ByRef value As Long )
			padbottom_ = value
		End Property

		Property Get OtherLayout() As iLayout
			Property = otherLayout_
		End Property
		Property Set OtherLayout( value As iLayout )
			otherLayout_ = value
			Call Me.SetHWnds()
		End Property

		Property Get SuspendPaint() As Long
			Property = suspendPaint_
		End Property
		Property Set SuspendPaint( ByVal value As Long )
			suspendPaint_ = IIf&( value, -1, 0 )
		End Property

		Property Get XGutter() As Long
			Property = xgutter_
		End Property
		Property Set XGutter( ByRef value As Long )
			If value>=0 Then xgutter_ = value
			Me.FlagForLayout()
		End Property

		Property Get YGutter() As Long
			Property = ygutter_
		End Property
		Property Set YGutter( ByRef value As Long )
			If value>=0 Then ygutter_ = value
			Me.FlagForLayout()
		End Property

		Property Get CurItem() As Long
			Property = curItem_
		End Property
		Property Set CurItem( ByVal value As Long )
			curItem_ = value
		End Property

		Property Get OtherItem() As Long
			Property = otherItem_
		End Property
		Property Set OtherItem( ByVal value As Long )
			otherItem_ = value
		End Property
		Property Get OtherItemCurrent() As Long
			Property = otherItemCurrent_
		End Property
		Property Set OtherItemCurrent( ByVal value As Long )
			otherItemCurrent_ = value
		End Property

		Method SetOtherBounds( ByVal start As Long, ByVal final As Long  )
			' Not used for flow
		End Method

		Property Get IsHotTracking() As Long
			Property = IIf& ( hotType_ <> %IVSelect.None, -1, 0 )
		End Property

		Property Get HWnd() As Dword
			Property = hWnd_
		End Property
		Property Set HWnd( ByVal value As Dword )
			hWnd_ = value
			Call Me.SetHWnds()
			Me.FlagForPaint()
		End Property

		Method GotoItem( ByVal offset As Long )
			curItem_ = offset
			If curItem_ < firstItem_ Then
				Me.FirstItem = curItem_
				Me.FlagForPaint()
			ElseIf curItem_ >= firstItem_ + pageSize_ Then
				Me.FirstItem = curItem_ - pageSize_ + 1
				Me.FlagForPaint()
			End If
		End Method

		Property Get SizeItem( ByVal offset As Long ) As Long
			If offset = 0 Then
				Property = width_
			Else
				Property = height_
			End If
		End Property

		Property Get SizeOverride( ByVal offset As Long ) As Long
			' Doesn't apply
		End Property
		Property Set SizeOverride( ByVal offset As Long, ByVal value As Long ): ' <0 = remove custom size
			' Doesn't apply
		End Property
		Method SizeOverrideClear( ByVal offset As Long )
			' Doesn't apply
		End Method

		Property Get FirstItem() As Long
			Property = firstItem_
		End Property
		Property Set FirstItem( value As Long )
			If firstItem_ <> value Then
				firstItem_ = Max&( 0, Min&( maxFirst_, value ) )
				Me.FlagForLayout()
			End If
		End Property

		Property Get MinFirst() As Long
			Property = 0
		End Property
		Property Get MaxFirst() As Long
			Property = maxFirst_
		End Property

		Property Get SelStart() As Long
			Property = selStart_
		End Property
		Property Set SelStart( ByVal value As Long )
			selStart_ = value
		End Property

		Property Get ItemCount() As Long
			If IsObject( dataSource_ ) Then
				If isCurRow_ Then
					Property = dataSource_.RowCount
				Else
					Property = dataSource_.ColCount
				End If
			End If
		End Property

		Property Get VisibleCount() As Long
			Property = visibleCount_
		End Property
		Property Get PageItemCount() As Long
			Property = visibleCount_
		End Property

		Property Get ItemIndex( ByVal offset As Long ) As Long
			If offset<0 Or offset>=visibleCount_ Then Exit Property
			Property = itemIndex_( offset)
		End Property

		Method GetFullBounds( rc As RECT )
			CopyRect( rc, bounds_ )
		End Method

		Method GetDrawnBounds( rc As RECT )
			Dim l As Long
			Dim t As Long
			Dim r As Long
			Dim b As Long
			Dim i As Long
			If visibleCount_ = 0 Then
				SetRect( rc, 0,0,0,0 )
			Else
				l = rc_(0).Left
				t = rc_(0).Top
				r = rc_(0).Right
				b = rc_(0).Bottom
				For i=1 To visibleCount_ - 1
					l = Min&( rc_(i).Left, l )
					t = Min&( rc_(i).Top, t )
					r = Max&( rc_(i).Right, r )
					b = Max&( rc_(i).Bottom, b )
				Next
				SetRect( rc, l, t, r, b )
			End If
		End Method

		Method GetRect( ByVal offset As Long, rc As RECT )
			If offset<0 Or offset>=visibleCount_ Then Exit Method
			CopyRect( rc, rc_( offset) )
		End Method

		Method GetBorder( ByVal x As Long, ByVal y As Long ) As Long
			' Doesn't apply
		End Method

		Method Hit( ByVal x As Long, ByVal y As Long ) As Long
			Local i As Long
			Local pt As Point
			Local rc As Rect
			pt.X = x
			pt.Y = y
			For i = 0 To visibleCount_-1
				CopyRect( rc, rc_( i ) )
				If rc.Top = ygutter_ Then rc.Top = 0
				If rc.Left = xgutter_ Then rc.Left = 0
				rc.Right += xgutter_
				rc.Bottom += ygutter_
				If PtInRect( rc, pt ) Then
					Method = i
					Exit Method
				End If
			Next
			Method = -1
		End Method

		Method EnsureLayoutUpdated( ByRef newBounds As RECT )
			Me.RequestLayoutUpdate( newBounds )
			If needLayout_ Then
				Me.DoUpdateLayout()
			End If
		End Method
		Method RequestLayoutUpdate( ByRef newBounds As RECT )
			If IsFalse EqualRect( bounds_, newBounds ) Then
				CopyRect( bounds_, newBounds )
				Me.FlagForLayout()
			End If
		End Method
	End Interface

	Interface iRender Guid$( "{c840a61d-d562-4b33-8d81-b243cbde20e5}")
		Inherit IUnknown

		Method Render( ByVal hdc As Dword, ByRef rc As RECT, ByVal r As Long, ByVal c As Long )
			Local i As Long
			Local rn As iRender
			Local cr As Long: ' Current row
			Local cc As Long: ' Current col
			Local st As Long
			Local fb As RECT
			Local crc As RECT
			Local cursor As Point
			Local v As WString
			Local seq As iFlowLayout
			Local lay As iLayout
			Local vc As Long
			Local o As Long
			Local sr As Long
			Local sc As Long
			Local sn As Long
			Local hClipRgn As Dword
			Local pRgn As Dword
			Local hPrevRgn As Dword

			seq = Me
			lay = Me

			sn = suspendPaint_
			suspendPaint_ = -1
			GetCursorPos( cursor)
			MapWindowPoints( %NULL, hwnd_, VarPtr( cursor), 1 )

			If needLayout_ Then Me.DoUpdateLayout()

			lay.GetDrawnBounds( fb ): ' Get full bounds

			hPrevRgn = CreateRectRgnIndirect( rc )
			If IsFalse GetClipRgn( hdc, hPrevRgn ) Then
				DeleteObject( hPrevRgn )
				hPrevRgn = 0
			End If
			hClipRgn = CreateRectRgnIndirect( rc )
			SelectClipRgn( hdc, hClipRgn )

			If isCurRow_ Then
				c = otherItem_
				cc = otherItem_
				cr = curItem_
				sr = selStart_
				sc = c
			Else
				r = otherItem_
				cr = otherItem_
				cc = curItem_
				sc = selStart_
				sr = r
			End If
			vc = visibleCount_ - 1
			For i = 0 To vc
				If isCurRow_ Then
					r = lay.ItemIndex( i )
					o = r
				Else
					c = lay.ItemIndex( i )
					o = c
				End If
				lay.GetRect( i, crc )
				st = %IVState.Normal
				If hotType_ = %IVSelect.Content And PtInRect( crc, cursor ) Then st = %IVState.Hot
				If selType_ = %IVSelect.Content And Between(r, sr, cr) Then st = %IVState.Selected
				If r < 0 Or c < 0 Then
					rn = seq.BackRenderer( -1, st )
				Else
					rn = seq.BackRenderer( o, st )
				End If
				Me.PaintPart( hdc, crc, r, c, "", %IVPart.Back, st, rn )

				If r < 0 Or c < 0 Then
					rn = seq.ContentRenderer( -1, st )
				Else
					rn = seq.ContentRenderer( o, st )
				End If
				v = ""
				If IsObject( dataSource_ ) Then v = dataSource_.Value( r, c )
				lay.GetRect( i, crc )
				Me.PaintPart( hdc, crc, r, c, v, %IVPart.Content, st, rn )

				If st = %IVState.Selected And GetFocus() = hWnd_ And r=cr Then
					lay.GetRect( i, crc )
					Me.PaintPart( hdc, crc, r, c, "", %IVPart.Focus, st, drawFocus_ )
				End If
			Next
			suspendPaint_ = 0
			If hPrevRgn Then
				SelectClipRgn( hdc, hPrevRgn )
				DeleteObject( hPrevRgn )
			End If
			DeleteObject( hClipRgn )
		End Method
	End Interface

	Interface iHasDatasource Guid$("{33e74a9a-2739-4180-a894-3fc9cd8614b3}")
		Inherit IUnknown

		Property Get DataSource() As iDataSource
			Property = dataSource_
		End Property
		Property Set DataSource( ByRef value As iDataSource )
			dataSource_ = value
			Me.FlagForPaint()
		End Property
	End Interface
End Class
