Function NewPositionedLayout() Common Export As iLayout
	Local obj As iLayout
	obj = Class "cPositionedLayout"
	Function = obj
End Function

Class cPositionedLayout Guid$("{c7dfe4f2-5e9a-477f-b6b8-afccc8b02b38}") Common
	Instance bounds_ As RECT:                         ' Bounds to render inside of
	Instance layout_ As RECT:                         ' Bounds layout made in
	Instance item_ As iPosList:                       ' List of positioned items
	Instance rc_() As RECT:                           ' Rectangle
	Instance itemIndex_() As Long:                    ' Column index
	Instance count_ As Long:                          ' # of rc_ or itemIndex_
	Instance drawBack_ As iLongStateRenderMap:        ' Item back renderer overrides ( back renderer for overridden backs)
	Instance drawContent_ As iLongStateRenderMap:     ' Item back renderer overrides ( back renderer for overridden backs)
	Instance alignLine_ As Long:                      ' Top, Middle, Bottom, Baseline (%IVAlign)
	Instance curItem_ As Long
	Instance otherItemCurrent_ As Long
	Instance hotType_ As Long
	Instance selType_ As Long

	Instance otherLayout_ As iLayout

	Instance suspendPaint_ As Long:                   ' <>0 = suspend painting
	Instance padLeft_ As Long
	Instance padRight_ As Long
	Instance padTop_ As Long
	Instance padBottom_ As Long

	Instance visibleCount_ As Long:                   ' # of visible items
	Instance otherItem_ As Long:                      ' other Item ( Current row or column #)
	Instance xgutter_ As Long:                        ' X Gutter size ( Distance between items on line )
	Instance ygutter_ As Long:                        ' Y Gutter size (Distance between lines )

	Instance dataSource_ As iDataSource:              ' Data source
	Instance hWnd_ As Dword:                          ' Window doing layout in

	Instance drawOverride_ As iUserRender:            ' User rendering override
	Instance needLayout_ As Long:                     ' <>0 if layout needed

	Class Method Create()
		item_ = NewPosList()
		alignLine_ = %IVAlign.Baseline
		drawContent_ = NewLongStateRenderMap()
		drawBack_ = NewLongStateRenderMap()
	End Method
	Class Method Destroy()
	End Method

	Class Method ExtendLayout()
		Local ub As Long
		ub = item_.Count
		ReDim rc_( ub )
		ReDim itemIndex_( ub )
	End Method

	Class Method PaintPart( ByVal hdc As Long, ByRef rc As RECT, ByVal r As Long, ByVal c As Long, ByRef value As WString, ByVal part As Long, ByVal state As Long, rn As iRender )
		Local userHandled As Long
		Local v As iValue
		Local l As iFlowLayout

		If IsInterface( drawOverride_, iUserRender ) Then
			userHandled = drawOverride_.Render( hdc, rc, r, c, value, part, State, rn )
		End If
		If IsFalse userHandled And IsInterface( rn, iRender ) Then
			v = rn: If IsObject( v ) Then v.Value = value
			rn.Render( hdc, rc, r, c )
		End If
	End Method

	Class Method FlagForPaint()
		If suspendPaint_ Or IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Then Exit Method
		InvalidateRect( hWnd_, bounds_, 0 )
	End Method

	Class Method FlagForLayout()
		needLayout_ = -1
		Me.FlagForPaint()
	End Method

	Class Method SetHWnds()
		Local key As Long
		Local i As Long
		Local srm As iStateRender

		If IsObject( otherLayout_ ) Then otherLayout_.HWnd = hWnd_

		For i=0 To drawBack_.Count-1
			key = drawBack_.Key(i)
			srm = drawBack_.Value(key)
			Call Me.SetStateRenderHWnd( srm )
		Next
		For i=0 To drawContent_.Count-1
			key = drawContent_.Key(i)
			srm = drawContent_.Value(key)
			Call Me.SetStateRenderHWnd( srm )
		Next
	End Method

	Class Method SetStateRenderHWnd( srm As iStateRender )
		Local rn As iRender
		If IsInterface( srm, iStateRender ) Then
			rn = srm.Item( %IVState.Normal ):   Call Me.SetRenderHWnd( rn )
			rn = srm.Item( %IVState.Selected ): Call Me.SetRenderHWnd( rn )
			rn = srm.Item( %IVState.Hot ):      Call Me.SetRenderHWnd( rn )
		End If
	End Method
	Class Method SetRenderHWnd( rn As iRender )
		Local lay As iLayout

		lay = rn
		If IsInterface( lay, iLayout ) Then lay.HWnd = hWnd_
	End Method

	Class Method DoUpdateLayout()
		Local i As Long
		Local rc As RECT
		Local anchor As Long
		Local ro As Long
		Local bo As Long

		Me.ExtendLayout()

		count_ = item_.Count

		For i=0 To count_-1
			item_.Item(i).Rect( rc )
			rc.Left += bounds_.Left
			rc.Right += bounds_.Left
			rc.Top += bounds_.Top
			rc.Bottom += bounds_.Top
			ro = (bounds_.Right - bounds_.Left) - (layout_.Right-layout_.Left)
			bo = (bounds_.Bottom - bounds_.Top) - (layout_.Bottom-layout_.Top)
			anchor = item_.Item(i).Anchor
			If (anchor And %IVAnchor.Right) Then
				rc.Right += ro
				If (anchor And %IVAnchor.Left)=0 Then
					rc.Left += ro
				End If
			End If
			If (anchor And %IVAnchor.Bottom) Then
				rc.Bottom += bo
				If (anchor And %IVAnchor.Top)=0 Then
					rc.Top += bo
				End If
			End If
			CopyRect( rc_(i), rc )
			itemIndex_(i) = item_.Item(i).Col
		Next
	End Method

	Class Method GetRenderMap( map As iLongStateRenderMap, ByVal offset As Long, ByVal state As Long ) As iRender
		Local rs As iStateRender
		Local rn As iRender
		If IsFalse IsObject( map ) Then Exit Method
		If map.Contains( offset ) Then
			rs = map.Value( offset )
			If IsObject( rs ) Then rn = rs.Item( State )
		End If
		If IsFalse IsObject( rn ) And map.Contains( %IVSpecial.Default ) Then
			rs = map.Value( %IVSpecial.Default )
			If IsObject( rs ) Then rn = rs.Item( State )
		End If
		Method = rn
	End Method

	Class Method GetMap( map As iLongStateRenderMap, ByVal offset As Long, ByVal state As Long ) As iRender
		Local rs As iStateRender

		If IsObject( map ) And State>=%IVState.Normal And State<=%IVState.Hot Then
			If map.Contains( offset ) Then
				rs = map.Value( offset )
				If IsObject( rs ) Then Method = rs.Item( State )
			End If
		End If
	End Method

	Class Method SetMap( map As iLongStateRenderMap, ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
		Local rs As iStateRender

		If IsObject( map ) And State>=%IVState.Normal And State<=%IVState.Hot Then
			If Not map.Contains( offset ) Then
				rs = NewStateRender()
				map.Add( offset, rs )
			Else
				rs = map.Value( offset )
				If Not IsObject( rs ) Then
					rs = NewStateRender()
					map.Value( offset ) = rs
				End If
			End If
			If IsObject( rs ) Then rs.Item( State ) = value
			Call Me.SetHWnds()
			Me.FlagForPaint()
		End If
	End Method

	Class Method ClearMap( map As iLongStateRenderMap, ByVal offset As Long )
		If IsObject( map ) Then map.Remove( offset )
	End Method

	Interface iPositionedLayout Guid$("{8d0f2b9d-343d-4dc9-ba30-94007c7ac657}")
		Inherit IUnknown

		Property Get Layout() As RECT
			Property = layout_
		End Property
		Property Set Layout( value As RECT )
			CopyRect( layout_, value )
		End Property

		Property Set Def( value As WString )
			Local lineCnt As Long
			Local lineNum As Long

			Local fieldParts() As String
			Local partCnt As Long
			Local newPos As iPosItem
			Local vCol As Long
			Local vLeft As Long
			Local vTop As Long
			Local vWidth As Long
			Local vHeight As Long
			Local vAnchor As Long

			lineCnt = ParseCount( value, $CrLf )
			Dim ln(lineCnt-1) As String
			Parse value, ln(), $CrLf
			item_.Clear()
			For lineNum = 0 To lineCnt-1
				partCnt = ParseCount( ln(lineNum), ":" )
				If partCnt= 6 Then
					ReDim fieldParts( partCnt-1 )
					Parse ln(lineNum), fieldParts(), ":"
					vCol = Val(fieldParts(0))
					vLeft = Val(fieldParts(1))
					vTop = Val(fieldParts(2))
					vWidth = Val(fieldParts(3))
					vHeight = Val(fieldParts(4))
					vAnchor = 0
					vAnchor Or= IIf&(InStr( LCase$(fieldParts(5)), "l")>0, %IVAnchor.Left, 0 )
					vAnchor Or= IIf&(InStr( LCase$(fieldParts(5)), "t")>0, %IVAnchor.Top, 0 )
					vAnchor Or= IIf&(InStr( LCase$(fieldParts(5)), "r")>0, %IVAnchor.Right, 0 )
					vAnchor Or= IIf&(InStr( LCase$(fieldParts(5)), "b")>0, %IVAnchor.Bottom, 0 )
					newPos = NewPosItem( vCol, vLeft, vTop, vWidth, vHeight, vAnchor )
					item_.Add( newPos )
				End If
			Next
			needLayout_ = -1
		End Property

		Property Get Item() As iPosList
			Property = item_
		End Property

		Property Get SelectType() As Long
			Property = selType_
		End Property
		Property Set SelectType( ByVal value As Long )
			Select Case value
				Case %IVSelect.None, %IVSelect.Content
					selType_ = value
			End Select
		End Property

		Property Get HotType() As Long
			Property = hotType_
		End Property
		Property Set HotType( ByVal value As Long )
			Select Case value
				Case %IVSelect.None, %IVSelect.Content
					hotType_ = value
			End Select
		End Property

		Property Get DrawBackRender( ByVal offset As Long, ByVal state As Long ) As iRender
			Property = Me.GetRenderMap( drawBack_, offset, State )
		End Property
		Property Get DrawBack( ByVal offset As Long, ByVal state As Long ) As iRender
			Property = Me.GetMap( drawBack_, offset, State )
		End Property
		Property Set DrawBack( ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
			Me.SetMap( drawBack_, offset, State, value )
		End Property
		Method DrawBackClear( ByVal offset As Long )
			Me.ClearMap( drawBack_, offset )
		End Method

		Property Get DrawContentRender( ByVal offset As Long, ByVal state As Long ) As iRender
			Property = Me.GetRenderMap( drawContent_, offset, State )
		End Property
		Property Get DrawContent( ByVal offset As Long, ByVal state As Long ) As iRender
			Property = Me.GetMap( drawContent_, offset, State )
		End Property
		Property Set DrawContent( ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
			Me.SetMap( drawContent_, offset, State, value )
		End Property
		Method DrawContentClear( ByVal offset As Long )
			Me.ClearMap( drawContent_, offset )
		End Method
	End Interface

	Interface iLayout Guid$( "{b87daa5f-0140-4b31-8bd1-ff4f0c56e666}")
		Inherit IUnknown

		Property Get PadLeft() As Long
			Property = padleft_
		End Property
		Property Set PadLeft( ByRef value As Long )
			padleft_ = value
		End Property

		Property Get PadTop() As Long
			Property = padtop_
		End Property
		Property Set PadTop( ByRef value As Long )
			padtop_ = value
		End Property

		Property Get PadRight() As Long
			Property = padright_
		End Property
		Property Set PadRight( ByRef value As Long )
			padright_ = value
		End Property

		Property Get PadBottom() As Long
			Property = padbottom_
		End Property
		Property Set PadBottom( ByRef value As Long )
			padbottom_ = value
		End Property

		Property Get OtherLayout() As iLayout
			Property = otherLayout_
		End Property
		Property Set OtherLayout( value As iLayout )
			otherLayout_ = value
		End Property

		Property Get SuspendPaint() As Long
			Property = suspendPaint_
		End Property
		Property Set SuspendPaint( ByVal value As Long )
			suspendPaint_ = IIf&( value, -1, 0 )
		End Property

		Property Get XGutter() As Long
			Property = xgutter_
		End Property
		Property Set XGutter( ByRef value As Long )
			If value>=0 And value<>xgutter_ Then
				xgutter_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get YGutter() As Long
			Property = ygutter_
		End Property
		Property Set YGutter( ByRef value As Long )
			If value>=0 And value<>ygutter_ Then
				ygutter_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get CurItem() As Long
			Property = curItem_
		End Property
		Property Set CurItem( ByVal value As Long )
			curItem_ = value
		End Property

		Property Get OtherItem() As Long
			Property = otherItem_
		End Property
		Property Set OtherItem( ByVal value As Long )
			otherItem_ = value
		End Property

		Property Get OtherItemCurrent() As Long
			Property = otherItemCurrent_
		End Property
		Property Set OtherItemCurrent( ByVal value As Long )
			otherItemCurrent_ = value
		End Property

		Method SetOtherBounds( ByVal start As Long, ByVal final As Long )
			' Doesn't apply
		End Method

		Property Get IsHotTracking() As Long
			'Doesn't apply
		End Property

		Property Get HWnd() As Dword
			Property = hWnd_
		End Property
		Property Set HWnd( ByVal value As Dword )
			hWnd_ = value
			Me.FlagForPaint()
		End Property

		Method GotoItem( ByVal offset As Long )
			curItem_ = offset
		End Method

		Property Get SizeItem( ByVal offset As Long ) As Long
			' Doesn't apply
		End Property

		Property Get SizeOverride( ByVal offset As Long ) As Long
			' Doesn't apply
		End Property

		Property Set SizeOverride( ByVal offset As Long, ByVal value As Long ): ' <0 = remove custom size
			' Doesn't apply
		End Property

		Method SizeOverrideClear( ByVal offset As Long )
			' Doesn't apply
		End Method

		Property Get FirstItem() As Long
			' Doesn't apply
		End Property
		Property Set FirstItem( value As Long )
			' Doesn't apply
		End Property

		Property Get MinFirst() As Long
			Property = 0
		End Property

		Property Get MaxFirst() As Long
			' Doesn't apply
		End Property

		Property Get SelStart() As Long
			' Nothing to do
		End Property

		Property Set SelStart( ByVal value As Long )
			' Nothing to do
		End Property

		Property Get ItemCount() As Long
			If IsObject( dataSource_ ) Then
				Property = dataSource_.ColCount
			End If
		End Property

		Property Get VisibleCount() As Long
			Property = visibleCount_
		End Property

		Property Get PageItemCount() As Long
			Property = visibleCount_
		End Property

		Property Get ItemIndex( ByVal offset As Long ) As Long
			If offset<0 Or offset>=visibleCount_ Then Exit Property
			Property = itemIndex_( offset)
		End Property

		Method GetFullBounds( rc As RECT )
			CopyRect( rc, bounds_ )
		End Method

		Method GetDrawnBounds( rc As RECT )
			Dim l As Long
			Dim t As Long
			Dim r As Long
			Dim b As Long
			Dim i As Long
			If visibleCount_ = 0 Then
				SetRect( rc, 0,0,0,0 )
			Else
				l = rc_(0).Left
				t = rc_(0).Top
				r = rc_(0).Right
				b = rc_(0).Bottom
				For i=1 To visibleCount_ - 1
					l = Min&( rc_(i).Left, l )
					t = Min&( rc_(i).Top, t )
					r = Max&( rc_(i).Right, r )
					b = Max&( rc_(i).Bottom, b )
				Next
				SetRect( rc, l, t, r, b )
			End If
		End Method

		Method GetRect( ByVal offset As Long, rc As RECT )
			If offset<0 Or offset>=visibleCount_ Then Exit Method
			CopyRect( rc, rc_( offset) )
		End Method

		Method GetBorder( ByVal x As Long, ByVal y As Long ) As Long
			Method = -1
		End Method

		Method Hit( ByVal x As Long, ByVal y As Long ) As Long
			Local i As Long
			Local pt As Point
			Local rc As Rect
			pt.X = x
			pt.Y = y
			For i = 0 To visibleCount_-1
				If PtInRect( rc_(i), pt ) Then
					Method = i
					Exit Method
				End If
			Next
			Method = -1
		End Method

		Method EnsureLayoutUpdated( ByRef newBounds As RECT )
			Me.RequestLayoutUpdate( newBounds )
			If needLayout_ Then
				Me.DoUpdateLayout()
			End If
		End Method

		Method RequestLayoutUpdate( ByRef newBounds As RECT )
			If IsFalse EqualRect( bounds_, newBounds ) Then
				CopyRect( bounds_, newBounds )
				Me.DoUpdateLayout()
			End If
		End Method
	End Interface

	Interface iRender Guid$( "{c840a61d-d562-4b33-8d81-b243cbde20e5}")
		Inherit IUnknown

		Method Render( ByVal hdc As Dword, ByRef rc As RECT, ByVal r As Long, ByVal c As Long )
			Local i As Long
			Local rn As iRender
			Local slay As iPositionedLayout
			Local lay As iLayout
			Local cursor As Point
			Local crc As RECT
			Local cl As Long
			Local v As WString
			Local st As Long
			Local rw As Long
			Local sr As Long
			Local cr As Long
			Local hClipRgn As Dword
			Local pRgn As Dword
			Local hPrevRgn As Dword

			slay = Me
			lay = Me

			suspendPaint_ = 1
			If needLayout_ Then lay.EnsureLayoutUpdated( rc )

			GetCursorPos( cursor)
			MapWindowPoints( %NULL, hwnd_, VarPtr( cursor), 1 )

			rw = otherItem_
			sr = -2
			cr = -2
			If IsObject( otherLayout_ ) Then
				sr = otherLayout_.SelStart
				cr = otherLayout_.CurItem
			End If

			hPrevRgn = CreateRectRgnIndirect( rc )
			If IsFalse GetClipRgn( hdc, hPrevRgn ) Then
				DeleteObject( hPrevRgn )
				hPrevRgn = 0
			End If
			hClipRgn = CreateRectRgnIndirect( rc )
			SelectClipRgn( hdc, hClipRgn )

			For i=0 To count_-1
				cl = itemIndex_(i)
				st = %IVState.Normal
				If hotType_ = %IVSelect.Content And PtInRect( rc, cursor ) Then st = %IVState.Hot
				If selType_ = %IVSelect.Content And Between(r, sr, cr) Then st = %IVState.Selected

				rn = slay.DrawBackRender(i, st)
				CopyRect( crc, rc_(i) )
				Me.PaintPart( hdc, crc, rw, cl, "", %IVPart.Back, st, rn )

				rn = slay.DrawContentRender(i, st)
				v = ""
				If IsObject( dataSource_ ) Then v = dataSource_.Value( r, c )
				CopyRect( crc, rc_(i) )
				Me.PaintPart( hdc, crc, rw, cl, v, %IVPart.Content, st, rn )
			Next
			suspendPaint_ = 0
			If hPrevRgn Then
				SelectClipRgn( hdc, hPrevRgn )
				DeleteObject( hPrevRgn )
			End If
			DeleteObject( hClipRgn )
		End Method
	End Interface

	Interface iHasDatasource Guid$("{33e74a9a-2739-4180-a894-3fc9cd8614b3}")
		Inherit IUnknown

		Property Get DataSource() As iDataSource
			Property = dataSource_
		End Property

		Property Set DataSource( ByRef value As iDataSource )
			dataSource_ = value
		End Property
	End Interface
End Class
