Function NewSequentialLayout() Common Export As iLayout
	Local o As iLayout
	o = Class "cSequentialLayout"
	Function = o
End Function

Class cSequentialLayout Guid$( "{cb2c6166-c2c7-4eaa-86f0-8b1c2bd67218}") Common
	Instance visibleCount_ As Long:                   ' # of visible items
	Instance curItem_ As Long:                        ' current Item ( Current column or row #)
	Instance selStart_ As Long:                       ' Selection start
	Instance firstItem_ As Long:                      ' First unfrozen item ( Column or Row #)
	Instance otherItem_ As Long:                      ' Other current item ( Current row or column #)
	Instance otherItemCurrent_ As Long:               ' Current other item (for focus)
	Instance otherLayout_ As iLayout:                 ' Other item layout
	Instance isCurRow_ As Long:                       ' -1 = curItem is row, 0 curItem is col
	Instance maxFirst_ As Long:                       ' Maximum index of firstItem
	Instance needLayout_ As Long:                     ' <>0 = need update
	Instance suspendPaint_ As Long                    ' <>0 = Suspend Flagging for paint
	Instance padLeft_ As Long
	Instance padRight_ As Long
	Instance padTop_ As Long
	Instance padBottom_ As Long

	Instance dataSource_ As iDataSource:              ' Data source
	Instance selType_ As Long:                        ' Selection is None, Cell, or Row
	Instance hotType_ As Long:                        ' Hot tracking is None, Cell, or Row
	Instance hWnd_ As Dword:                          ' Window doing layout in

	Instance bounds_ As RECT:                         ' Bounds to render inside of
	Instance rc_() As RECT:                           ' Locations of each item
	Instance itemIndex_() As Long:                    ' Offsets of items ( Column or Row #'s)

	Instance gutter_ As Long                          ' Gutter size ( Distance between items, i.e. Grid )
	Instance frozen_ As Long:                         ' # of frozen items ( Frozen rows or columns)
	Instance horizontal_ As Long:                     ' 1 = horizontal, 0 = vertical

	Instance otherSize_ As Long:                      ' Other size direction ( Width or Height)
	Instance sizeDefault_ As Long:                    ' Item size ( Height or Width)
	Instance sizeOverride_ As iLongLongMap:           ' item size overrides ( Height or Width)

	Instance alignDefault_ As Long:                   ' Default alignment
	Instance alignOverride_ As iLongLongMap:           ' Alignment overrides

	Instance backDefault_ As iStateRender:            ' Item back renderers ( back renderer for non-overridden backs)
	Instance backOverride_ As iLongStateRenderMap:    ' Item back renderer overrides ( back renderer for overridden backs)

	Instance contentDefault_ As iStateRender:         ' Content renderers ( content renderer for non-overridden contents)
	Instance contentOverride_ As iLongStateRenderMap: ' Item content renderer overrides ( content renderer for overridden contents)

	Instance drawFocus_ As iRender:                   ' Draw focus rect

	Instance drawOverride_ As iUserRender

	Class Method Create()
		Dim rc_( 100)
		Dim itemIndex_( 100)

		alignDefault_ = %IVAlign.Left Or %IVAlign.Top

		sizeOverride_ = NewLongLongMap()
		alignOverride_ = NewLongLongMap()

		backDefault_ = NewStateRender()
		contentDefault_ = NewStateRender()
		backOverride_ = NewLongStateRenderMap()
		contentOverride_ = NewLongStateRenderMap()
	End Method

	Class Method SetHWnds()
		Local key As Long
		Local i As Long
		Local srm As iStateRender

		If IsObject( otherLayout_ ) Then otherLayout_.HWnd = hWnd_

		For i=0 To backOverride_.Count-1
			key = backOverride_.Key(i)
			srm = backOverride_.Value(key)
			Call Me.SetStateRenderHWnd( srm )
		Next
		For i=0 To contentOverride_.Count-1
			key = contentOverride_.Key(i)
			srm = contentOverride_.Value(key)
			Call Me.SetStateRenderHWnd( srm )
		Next

		Call Me.SetStateRenderHWnd( backDefault_ )
		Call Me.SetStateRenderHWnd( contentDefault_ )
		Call Me.SetRenderHWnd( drawFocus_ )
	End Method

	Class Method SetStateRenderHWnd( srm As iStateRender )
		Local rn As iRender
		If IsInterface( srm, iStateRender ) Then
			rn = srm.Item( %IVState.Normal ):   Call Me.SetRenderHWnd( rn )
			rn = srm.Item( %IVState.Selected ): Call Me.SetRenderHWnd( rn )
			rn = srm.Item( %IVState.Hot ):      Call Me.SetRenderHWnd( rn )
		End If
	End Method
	Class Method SetRenderHWnd( rn As iRender )
		Local lay As iLayout

		lay = rn
		If IsInterface( lay, iLayout ) Then lay.HWnd = hWnd_
	End Method

	Class Method ExtendLayout( ByRef ub As Long )
		ub += 100
		ReDim Preserve rc_( ub )
		ReDim Preserve itemIndex_( ub )
	End Method

	Class Method PaintPart( ByVal hdc As Long, ByRef rc As RECT, ByVal r As Long, ByVal c As Long, ByRef value As WString, ByVal part As Long, ByVal state As Long, rn As iRender )
		Local userHandled As Long
		Local v As iValue
		Local a As iAlign
		Local seq As iSequentialLayout
		Local lay As iLayout
		Local sn As Long

		If IsInterface( drawOverride_, iUserRender ) Then
			userHandled = drawOverride_.Render( hdc, rc, r, c, value, part, State, rn )
		End If
		If IsFalse userHandled And IsInterface( rn, iRender ) Then
			v = rn: If IsObject( v ) Then v.Value = value
			a = rn:
			If IsObject( a ) Then
				seq = Me
				If isCurRow_ Then
					a.Align = seq.GetAlign( r )
				Else
					If r<0 Then
						a.Align = seq.GetAlign( -1 )
					Else
						a.Align = seq.GetAlign( c )
					End If
				End If
			End If
			lay = rn
			If IsInterface( lay, iLayout ) Then
				If isCurRow_ Then
					lay.OtherItem = c
				Else
					lay.OtherItem = r
				End If
				lay.RequestLayoutUpdate( rc )
				sn = lay.SuspendPaint
				lay.SuspendPaint=1
				lay.EnsureLayoutUpdated( rc )
				lay.SetOtherBounds( rc.Left, rc.Right )
				lay.SuspendPaint = sn
			End If
			rn.Render( hdc, rc, r, c )
		Else
			rn = rn
		End If
	End Method

	Class Method FlagForPaint()
		If IsFalse IsWindow( hWnd_ ) Or IsFalse IsWindowVisible( hWnd_ ) Or suspendPaint_ Then Exit Method
		InvalidateRect( hWnd_, ByVal %NULL, 0 ): 'bounds_, 0 )
	End Method
	Class Method FlagForLayout()
		Local lay As iLayout
		Local srn As iStateRender
		Local rn As iRender

		needLayout_ = -1

		Me.FlagForPaint()
	End Method
	Class Method IsRowSelection_() As Long
		Local oseq As iSequentialLayout
		oseq = otherLayout_
		Method = IIf&( IsTrue IsObject( oseq ) And oseq.SelectType = %IVSelect.Content And horizontal_, -1, 0 )
	End Method

	Class Method DoUpdateLayout()
		Local i As Long
		Local e As Long
		Local x As Long
		Local y As Long
		Local w As Long
		Local h As Long
		Local ub As Long
		Local seq As iSequentialLayout
		Local lay As iLayout

		If bounds_.Left = bounds_.Right Or bounds_.Top = bounds_.Bottom Then Exit Method
		needLayout_ = 0
		seq = Me
		lay = Me
		x = bounds_.Left + padLeft_
		y = bounds_.Top + padTop_
		ub = UBound( rc_() )
		visibleCount_ = 0
		If horizontal_ Then
			h = otherSize_
			' Add visible headers and frozen cells
			For i=-1 To frozen_-1
				If x > bounds_.Right-padRight_ Then Exit For
				If visibleCount_ > ub Then Me.ExtendLayout( ub )
				w = lay.SizeItem( i )
				If w>0 Then
					itemIndex_( visibleCount_ ) = i
					SetRect( rc_( visibleCount_ ), x, y, x + w, y + h )
					Incr visibleCount_
					x += gutter_ + w
				End If
			Next
			For i=firstItem_ To lay.ItemCount-1
				If x > bounds_.Right-padRight_ Then Exit For
				If visibleCount_ > ub Then Me.ExtendLayout( ub )
				w = lay.SizeItem( i )
				If w > 0 Then
					itemIndex_( visibleCount_ ) = i
					SetRect( rc_( visibleCount_ ), x, y, x + w, y + h )
					Incr visibleCount_
					x += gutter_ + w
				End If
			Next
			' Find max value for FirstItem
			e = bounds_.Right - bounds_.Left - padLeft_ - padRight_: ' Width
			For i = -1 To frozen_ - 1
				x = lay.SizeItem( i )
				If x>0 Then e -= ( x + gutter_ ): ' Minus frozen/gutter
			Next
			w = 0
			For i=lay.ItemCount-1 To 0 Step -1
				x = lay.SizeItem( i )
				If x>0 Then w += x + gutter_
				If w>e Then
					' This one took us past edge
					Incr i
					maxFirst_ = Max&( frozen_, Min&( i, lay.ItemCount-1 ) )
					lay.FirstItem = firstItem_: ' In case it needs adjusting now
					Exit For
				End If
			Next
		Else
			' Vertical
			w = otherSize_
			For i=-1 To frozen_-1
				If y > bounds_.Bottom - padBottom_ Then Exit For
				If visibleCount_ > ub Then Me.ExtendLayout( ub )
				h = lay.SizeItem( i )
				If h>0 Then
					itemIndex_( visibleCount_ ) = i
					SetRect( rc_( visibleCount_), x, y, x + w, y + h )
					Incr visibleCount_
					y += gutter_ + h
				End If
			Next
			For i = firstItem_ To lay.ItemCount-1
				If y > bounds_.Bottom - padBottom_ Then Exit For
				If visibleCount_ > ub Then Me.ExtendLayout( ub )
				h = lay.SizeItem( i )
				If h>0 Then
					itemIndex_( visibleCount_ ) = i
					SetRect( rc_( visibleCount_), x, y, x + w, y + h )
					Incr visibleCount_
					y += gutter_ + h
				End If
			Next
			' Find max value for FirstItem
			e = bounds_.Bottom - bounds_.Top - padBottom_ - padTop_
			For i = -1 To frozen_ - 1
				y = lay.SizeItem( i )
				If y>0 Then e -= ( y + gutter_ )
			Next
			h = 0
			For i=lay.ItemCount-1 To 0 Step -1
				y = lay.SizeItem( i )
				If y>0 Then h += y + gutter_
				If h>e Then
					Incr i
					maxFirst_ = Max( frozen_, Min&( i, lay.ItemCount-1 ) )
					lay.FirstItem = firstItem_
					Exit For
				End If
			Next
		End If
	End Method

	Interface iSequentialLayout Guid$( "{59884037-f858-4957-a7ad-f7a2a17354e5}")
		Inherit IUnknown

		Property Get IsRowSelection() As Long
			Property = Me.IsRowSelection_()
		End Property

		Property Get IsHorizontal() As Long
			Property = horizontal_
		End Property
		Property Set IsHorizontal( ByVal value As Long )
			value = IIf&( IsTrue value, -1, 0 )
			If value<>horizontal_ Then
				horizontal_ = value
				Me.FlagForLayout()
			End If
		End Property

		Method GetAlign( ByVal offset As Long ) As Long
			If alignOverride_.Contains( offset ) Then
				Method = alignOverride_.Value( offset )
			Else
				Method = alignDefault_
			End If
		End Method

		Property Get DefaultAlign() As Long
			Property = alignDefault_
		End Property
		Property Set DefaultAlign( ByRef value As Long )
			alignDefault_ = value
			Me.FlagForPaint()
		End Property

		Property Get AlignOverride( ByVal offset As Long ) As Long
			If alignOverride_.Contains( offset ) Then
				Property = alignOverride_.Value( offset)
			Else
				Property = %IVAlign.None
			End If
		End Property
		Property Set AlignOverride( ByVal offset As Long, ByRef value As Long )
			If IsFalse alignOverride_.Contains( offset ) Then
				alignOverride_.Add( offset, value )
			Else
				alignOverride_.Value( offset ) = value
			End If
			Me.FlagForPaint()
		End Property
		Method ClearAlignOverride( ByVal offset As Long )
			If alignOverride_.Contains( offset ) Then
				alignOverride_.Remove( offset )
				Me.FlagForPaint()
			End If
		End Method

		Property Get IsCurrentItemARow() As Long
			Property = isCurRow_
		End Property
		Property Set IsCurrentItemARow( ByVal value As Long )
			If value <> isCurRow_ Then
				isCurRow_ = IIf&( value, -1, 0 )
				Me.FlagForPaint()
			End If
		End Property

		Property Get SelectType() As Long
			Property = selType_
		End Property
		Property Set SelectType( ByVal value As Long )
			Select Case value
				Case %IVSelect.None, %IVSelect.Content
					selType_ = value
					Me.FlagForPaint()
			End Select
		End Property

		Property Get HotType() As Long
			Property = hotType_
		End Property
		Property Set HotType( ByVal value As Long )
			Select Case value
				Case %IVSelect.None, %IVSelect.Content
					hotType_ = value
					Me.FlagForPaint()
			End Select
		End Property

		Property Get Frozen() As Long
			Property = frozen_
		End Property
		Property Set Frozen( ByVal value As Long )
			value = Max&( 0, value )
			If frozen_<>value Then
				frozen_ = Max&( 0, value ): 'Min&( l.ItemCount-1, value ) )
				Me.FlagForLayout()
			End If
		End Property

		Property Get BackRenderer( ByVal offset As Long, ByVal state As Long ) As iRender
			Local r As iStateRender
			Local rn As iRender
			r = backOverride_.Value( offset )
			If IsObject( r ) Then
				rn = r.Item( State )
				If IsFalse IsObject( rn ) Then
					r = Nothing
				End If
			End If
			If IsFalse IsObject( r ) Then
				r = backDefault_
			End If
			If IsObject( r ) Then
				Property = r.Item( State )
			End If
		End Property
		Property Get ContentRenderer( ByVal offset As Long, ByVal state As Long ) As iRender
			Local r As iStateRender
			Local rn As iRender
			r = contentOverride_.Value( offset )
			If IsObject( r ) Then
				rn = r.Item( State )
				If IsFalse IsObject( rn ) Then
					r = Nothing
				End If
			End If
			If IsFalse IsObject( r ) Then
				r = contentDefault_
			End If
			If IsObject( r ) Then
				rn = r.Item( State )
				If IsObject( rn ) Then
					Property = rn
				End If
			End If
		End Property

		Property Get DrawFocus() As iRender
			Property = drawFocus_
		End Property
		Property Set DrawFocus( ByRef value As iRender )
			drawFocus_ = value
			Me.FlagForPaint()
		End Property

		Property Get DrawOverride() As iUserRender
			Property = drawOverride_
		End Property
		Property Set DrawOverride( value As iUserRender )
			drawOverride_ = value
			Me.FlagForPaint()
		End Property

		Property Get BackDefault( ByVal state As Long ) As iRender
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				Property = backDefault_.Item( State )
			End If
		End Property
		Property Set BackDefault( ByVal state As Long, ByRef value As iRender )
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				backDefault_.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property

		Property Get BackOverride( ByVal offset As Long, ByVal state As Long ) As iRender
			Local rs As iStateRender
			Local l As iLayout
			l = Me
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				If backOverride_.Contains( offset ) Then
					rs = backOverride_.Value( offset )
					If IsObject( rs ) Then Property = rs.Item( State )
				End If
			End If
		End Property
		Property Set BackOverride( ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
			Local rs As iStateRender
			Local l As iLayout
			l = Me
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				If Not backOverride_.Contains( offset ) Then
					rs = NewStateRender()
					backOverride_.Add( offset, rs )
				Else
					rs = backOverride_.Value( offset )
				End If
				If IsObject( rs ) Then rs.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property
		Method BackOverrideClear( ByVal offset As Long )
			backOverride_.Remove( offset )
		End Method

		Property Get ContentDefault( ByVal state As Long ) As iRender
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				Property = contentDefault_.Item( State )
			End If
		End Property
		Property Set ContentDefault( ByVal state As Long, ByRef value As iRender )
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				contentDefault_.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property

		Property Get ContentOverride( ByVal offset As Long, ByVal state As Long ) As iRender
			Local rs As iStateRender
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				If contentOverride_.Contains( offset ) Then
					rs = contentOverride_.Value( offset )
					If IsObject( rs ) Then Property = rs.Item( State )
				End If
			End If
		End Property
		Property Set ContentOverride( ByVal offset As Long, ByVal state As Long, ByRef value As iRender )
			Local rs As iStateRender
			If State>=%IVState.Normal And State<=%IVState.Hot Then
				If Not contentOverride_.Contains( offset ) Then
					rs = NewStateRender()
					contentOverride_.Add( offset, rs )
				Else
					rs = contentOverride_.Value( offset )
				End If
				If IsObject( rs ) Then rs.Item( State ) = value
				Call Me.SetHWnds()
				Me.FlagForPaint()
			End If
		End Property

		Method ContentOverrideClear( ByVal offset As Long )
			contentOverride_.Remove( offset )
			Me.FlagForPaint()
		End Method

		Property Get Gutter() As Long
			Property = gutter_
		End Property
		Property Set Gutter( ByRef value As Long )
			value = Max&( 0, value )
			If gutter_ <> value Then
				gutter_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get SizeDefault() As Long
			Property = sizeDefault_
		End Property
		Property Set SizeDefault( ByRef value As Long )
			If sizeDefault_ <> value Then
				sizeDefault_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get OtherSize() As Long
			Property = othersize_
		End Property
		Property Set OtherSize( ByRef value As Long )
			If otherSize_ <> value Then
				othersize_ = value
				Me.FlagForLayout()
			End If
		End Property

		Method GetOtherBounds( ByRef start As Long, ByRef final As Long )
			If horizontal_ Then
				start = rc_(0).Top
				final = rc_(0).Bottom
			Else
				start = rc_(0).Left
				final = rc_(0).Right
			End If
		End Method
	End Interface

	Interface iLayout Guid$( "{b87daa5f-0140-4b31-8bd1-ff4f0c56e666}")
		Inherit IUnknown

		Property Get PadLeft() As Long
			Property = padleft_
		End Property
		Property Set PadLeft( ByRef value As Long )
			padleft_ = value
		End Property

		Property Get PadTop() As Long
			Property = padtop_
		End Property
		Property Set PadTop( ByRef value As Long )
			padtop_ = value
		End Property

		Property Get PadRight() As Long
			Property = padright_
		End Property
		Property Set PadRight( ByRef value As Long )
			padright_ = value
		End Property

		Property Get PadBottom() As Long
			Property = padbottom_
		End Property
		Property Set PadBottom( ByRef value As Long )
			padbottom_ = value
		End Property

		Property Get OtherLayout() As iLayout
			Property = otherLayout_
		End Property
		Property Set OtherLayout( value As iLayout )
			otherLayout_ = value
			Call Me.SetHWnds()
		End Property

		Property Get SuspendPaint() As Long
			Property = suspendPaint_
		End Property
		Property Set SuspendPaint( ByVal value As Long )
			suspendPaint_ = IIf&( value, -1, 0 )
		End Property

		Property Get XGutter() As Long
			Property = gutter_
		End Property
		Property Set XGutter( ByRef value As Long )
			value= Max&( 0, value )
			If value<>gutter_ Then
				gutter_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get YGutter() As Long
			Property = gutter_
		End Property
		Property Set YGutter( ByRef value As Long )
			Me.XGutter = value
		End Property

		Property Get CurItem() As Long
			Property = curItem_
		End Property
		Property Set CurItem( ByVal value As Long )
			If value <> curItem_ Then
				curItem_ = value
				Me.FlagForPaint(): ' Changing cur item only means possible redraw?
			End If
		End Property

		Property Get OtherItem() As Long
			Property = otherItem_
		End Property
		Property Set OtherItem( ByVal value As Long )
			If value<>otherItem_ Then
				otherItem_ = value
				Me.FlagForPaint()
			End If
		End Property

		Property Get OtherItemCurrent() As Long
			Property = otherItemCurrent_
		End Property
		Property Set OtherItemCurrent( ByVal value As Long )
			otherItemCurrent_ = value
		End Property

		Method SetOtherBounds( ByVal start As Long, ByVal final As Long  )
			' Quick update of other dimension, assumes all values on linear axis
			' Useful for grids/tables when painting
			Local i As Long

			If horizontal_ Then
				For i=0 To visibleCount_ - 1
					rc_( i).Top = start + padTop_
					rc_( i).Bottom = final - padBottom_
				Next
			Else
				For i=0 To visibleCount_ - 1
					rc_( i).Left = start + padLeft_
					rc_( i).Right = final - padRight_
				Next
			End If
		End Method

		Property Get IsHotTracking() As Long
			Property = IIf& ( hotType_ <> %IVSelect.None, -1, 0 )
		End Property

		Property Get HWnd() As Dword
			Property = hWnd_
		End Property
		Property Set HWnd( ByVal value As Dword )
			hWnd_ = value
			Call Me.SetHWnds()
			Me.FlagForPaint()
		End Property

		Method GotoItem( ByVal offset As Long )
			Local w As Long
			Local e As Long
			Local i As Long
			Local fi As Long
			Local seq As iSequentialLayout
			Local lay As iLayout

			seq = Me
			lay = Me
			Me.EnsureLayoutUpdated( bounds_ )
			Me.CurItem = Max&( 0, Min&( Me.ItemCount-1, offset ) )

			If curItem_ < firstItem_ Then
				lay.FirstItem = curItem_
			ElseIf visibleCount_>0 And curItem_ >= itemIndex_( visibleCount_ - 1 ) Then
				fi = curItem_
				' Remove headers sizes
				w = lay.SizeItem( -1 ) + gutter_
				' Remove frozen sizes
				For i = 0 To frozen_-1
					w += lay.SizeItem( i ) + gutter_
				Next
				If horizontal_ Then
					e = bounds_.Right
				Else
					e = bounds_.Bottom
				End If
				w += lay.SizeItem( fi ) + gutter_
				While w <= e
					Decr fi
					w += lay.SizeItem( fi ) + gutter_
				Wend
				Me.FirstItem = fi+1
			End If
		End Method

		Property Get SizeItem( ByVal offset As Long ) As Long
			If sizeOverride_.Contains( offset ) Then
				Property = sizeOverride_.Value( offset )
			Else
				Property = sizeDefault_
			End If
		End Property

		Property Get SizeOverride( ByVal offset As Long ) As Long
			If sizeOverride_.Contains( offset ) Then
				Property = sizeOverride_.Value( offset )
			Else
				Property = 0
			End If
		End Property
		Property Set SizeOverride( ByVal offset As Long, ByVal value As Long )
			If value<=0 Then
				value = 0
			Else
				value = Max&( 3, value )
			End If

			If sizeOverride_.Contains( offset) Then
				If sizeOverride_.Value( offset ) <> value Then
					sizeOverride_.Value( offset) = Max&( 3, value )
					Me.FlagForLayout()
				End If
			Else
				sizeOverride_.Add( offset, value )
				Me.FlagForLayout()
			End If
		End Property
		Method SizeOverrideClear( ByVal offset As Long )
			If sizeOverride_.Contains( offset ) Then
				sizeOverride_.Remove( offset )
				Me.FlagForLayout()
			End If
		End Method

		Property Get FirstItem() As Long
			Property = firstItem_
		End Property
		Property Set FirstItem( value As Long )
			value = Max&( frozen_, Min&( maxFirst_, value ) )
			If firstItem_ <> value Then
				firstItem_ = value
				Me.FlagForLayout()
			End If
		End Property

		Property Get MinFirst() As Long
			Property = frozen_
		End Property
		Property Get MaxFirst() As Long
			Property = maxFirst_
		End Property

		Property Get SelStart() As Long
			Property = selStart_
		End Property
		Property Set SelStart( ByVal value As Long )
			selStart_ = value
			Me.FlagForPaint()
		End Property

		Property Get ItemCount() As Long
			If IsObject( dataSource_ ) Then
				If isCurRow_ Then
					Property = dataSource_.RowCount
				Else
					Property = dataSource_.ColCount
				End If
			End If
		End Property

		Property Get VisibleCount() As Long
			Property = visibleCount_
		End Property
		Property Get PageItemCount() As Long
			Dim i As Long
			Dim c As Long
			Local seq As iSequentialLayout
			seq = Me
			c = visibleCount_
			For i=-1 To frozen_
				If Me.SizeItem( i ) > 0 Then Decr c
			Next
			Property = Max&( 1, c )
		End Property

		Property Get ItemIndex( ByVal offset As Long ) As Long
			If offset<0 Or offset>=visibleCount_ Then Exit Property
			Property = itemIndex_( offset)
		End Property

		Method GetFullBounds( rc As RECT )
			CopyRect( rc, bounds_ )
		End Method

		Method GetDrawnBounds( rc As RECT )
			If visibleCount_ = 0 Then
				SetRect( rc, 0,0,0,0 )
			Else
				SetRect( rc, rc_( 0).Left, rc_( 0).Top, rc_( visibleCount_-1).Right, rc_( visibleCount_-1).Bottom )
			End If
		End Method

		Method GetRect( ByVal offset As Long, rc As RECT )
			If offset<0 Or offset>=visibleCount_ Then Exit Method
			CopyRect( rc, rc_( offset) )
		End Method

		Method GetBorder( ByVal x As Long, ByVal y As Long ) As Long
			Local i As Long
			Local a1 As Long
			Local a2 As Long

			For i = 0 To visibleCount_ - 1
				If horizontal_ Then
					a1 = rc_( i ).Right - 2
					If i < visibleCount_ - 1 Then
						a2 = rc_( i+1 ).Left + 2
					Else
						a2 = rc_( i ).Right + 2
					End If
					If x>=a1 And x<=a2 Then
						Method = i
						Exit Method
					End If
				Else
					a1 = rc_( i ).Bottom - 2
					If i < visibleCount_ - 1 Then
						a2 = rc_( i+1 ).Top + 2
					Else
						a2 = rc_( i ).Bottom + 2
					End If
					If y>=a1 And y<=a2 Then
						Method = i
						Exit Method
					End If
				End If
			Next
			Method = -1
		End Method

		Method Hit( ByVal x As Long, ByVal y As Long ) As Long
			Local i As Long
			Local pt As Point
			Local rc As Rect
			pt.X = x
			pt.Y = y
			For i = 0 To visibleCount_-1
				CopyRect( rc, rc_( i ) )
				If horizontal_ Then
					rc.Right += gutter_
				Else
					rc.Bottom += gutter_
				End If
				If PtInRect( rc, pt ) Then
					Method = i
					Exit Method
				End If
			Next
			Method = -1
		End Method
		Method EnsureLayoutUpdated( ByRef newBounds As RECT )
			Me.RequestLayoutUpdate( newBounds )
			If needLayout_ Then
				Me.DoUpdateLayout()
			End If
		End Method
		Method RequestLayoutUpdate( ByRef newBounds As RECT )
			If EqualRect( bounds_, newBounds )=0 Then
				CopyRect( bounds_, newBounds )
				Me.FlagForLayout()
			End If
		End Method
	End Interface

	Interface iRender Guid$( "{c840a61d-d562-4b33-8d81-b243cbde20e5}")
		Inherit IUnknown

		Method Render( ByVal hdc As Dword, ByRef rc As RECT, ByVal r As Long, ByVal c As Long )
			Local i As Long
			Local cr As Long: ' Current row
			Local cc As Long: ' Current col
			Local st As Long
			Local fb As RECT
			Local cursor As Point
			Local v As WString
			Local crc As RECT
			Local vc As Long
			Local itmIdx As Long
			Local sr As Long
			Local sc As Long
			Local rn As iRender
			Local lay As iLayout
			Local seq As iSequentialLayout
			Local oseq As iSequentialLayout
			Local x As Long
			Local hClipRgn As Dword
			Local pRgn As Dword
			Local hPrevRgn As Dword

			If IsFalse IsWindow( hWnd_ ) Then Exit Method
			oseq = otherLayout_
			seq = Me
			lay = Me

			If needLayout_ Then Call Me.DoUpdateLayout()
			GetCursorPos( cursor)
			MapWindowPoints( %NULL, hwnd_, VarPtr( cursor), 1 )

			lay.GetDrawnBounds( fb ): ' Get full bounds

			hPrevRgn = CreateRectRgnIndirect( rc )
			If IsFalse GetClipRgn( hdc, hPrevRgn ) Then
				DeleteObject( hPrevRgn )
				hPrevRgn = 0
			End If
			hClipRgn = CreateRectRgnIndirect( rc )
			SelectClipRgn( hdc, hClipRgn )

			If isCurRow_ Then
				c = otherItem_
				cc = otherItemCurrent_
				cr = curItem_
				sr = selStart_
				sc = c
			Else
				r = otherItem_
				cr = otherItemCurrent_
				cc = curItem_
				sc = selStart_
				sr = r
			End If
			vc = visibleCount_ - 1
			For i = 0 To vc
				If isCurRow_ Then
					r = lay.ItemIndex( i )
					itmIdx = r
				Else
					c = lay.ItemIndex( i )
					itmIdx = c
				End If
				lay.GetRect( i, crc )
				st = %IVState.Normal
				If hotType_ = %IVSelect.Content And PtInRect( crc, cursor ) Then st = %IVState.Hot
				If selType_ = %IVSelect.Content And Between(r, sr, cr) And Between(c, sc, cc ) Then st = %IVState.Selected
				If selType_ = %IVSelect.Content And horizontal_ Then
					If crc.Top<=cursor.y And crc.Bottom>=cursor.y And r>-1 Then
						rn = Nothing
						If IsObject( oseq) Then rn = oseq.BackRenderer( r, %IVState.Hot )
						If IsObject( rn ) Then st = %IVState.Hot
					End If
					If Between( r, otherLayout_.SelStart, otherLayout_.CurItem ) Then
						rn = Nothing
						If IsObject( oseq) Then rn = oseq.BackRenderer( r, %IVState.Selected )
						If IsObject( rn ) Then st = %IVState.Selected
					End If
					If c=-1 And Between( r, otherLayout_.SelStart, otherLayout_.CurItem ) Then st = %IVState.Hot
					If r=-1 And Between( c, sc, cc ) Then st = %IVState.Hot
				End If
				rn = Nothing
				x=-2
				If horizontal_ Then
					If r < 0 Then
						If IsObject(oseq) Then rn = oseq.BackRenderer( -1, st )
						If IsFalse IsObject( rn ) Then rn = seq.BackRenderer( -1, st )
					ElseIf c<0 Then
						rn = seq.BackRenderer( -1, st )
					ElseIf IsInterface( oseq, iSequentialLayout ) And IsInterface( otherLayout_, iLayout) And Between( r, otherLayout_.SelStart, otherLayout_.CurItem ) And st<>%IVState.Normal Then
						rn = oseq.BackRenderer( r, st )
					ElseIf IsInterface( oseq, iSequentialLayout ) And oseq.Frozen > r Then
						rn = oseq.BackRenderer( r, st )
					End If
					If IsFalse IsObject( rn ) Then
						rn = seq.BackRenderer( itmIdx, st )
					End If
				Else
					If otherItem_ = -1 Then
						rn = seq.BackRenderer( -1, st )
					Else
						rn = seq.BackRenderer( itmIdx, st )
					End If
				End If
				Me.PaintPart( hdc, crc, r, c, "", %IVPart.Back, st, rn )
				rn = Nothing
				If horizontal_ Then
					If r < 0 Then
						If IsObject(oseq) Then rn = oseq.ContentRenderer( -1, st )
						If IsFalse IsObject( rn ) Then rn = seq.ContentRenderer( -1, st )
					ElseIf c<0 Then
						rn = seq.ContentRenderer( -1, st )
					ElseIf IsInterface( oseq, iSequentialLayout ) And oseq.Frozen > r Then
						rn = oseq.ContentRenderer( r, st )
					End If
					If IsFalse IsObject( rn ) Then
						rn = seq.ContentRenderer( itmIdx, st )
					End If
				Else
					If otherItem_ = -1 Then
						rn = seq.ContentRenderer( -1, %IVState.Normal )
						x = otherItem_
					Else
						rn = seq.ContentRenderer( itmIdx, st )
					End If
				End If
				v = ""
				If IsObject( dataSource_ ) Then v = dataSource_.Value( r, c )
				lay.GetRect( i, crc )
				If x>-2 Then c=x
				Me.PaintPart( hdc, crc, r, c, v, %IVPart.Content, st, rn )

				If st = %IVState.Selected And GetFocus() = hWnd_ And r=cr And c=cc Then
					lay.GetRect( i, crc )
					Me.PaintPart( hdc, crc, r, c, "", %IVPart.Focus, st, drawFocus_ )
				End If
			Next
			If hPrevRgn Then
				SelectClipRgn( hdc, hPrevRgn )
				DeleteObject( hPrevRgn )
			End If
			DeleteObject( hClipRgn )
		End Method
	End Interface

	Interface iHasDatasource Guid$("{33e74a9a-2739-4180-a894-3fc9cd8614b3}")
		Inherit IUnknown

		Property Get DataSource() As iDataSource
			Property = dataSource_
		End Property
		Property Set DataSource( ByRef value As iDataSource )
			dataSource_ = value
			Me.FlagForLayout()
		End Property
	End Interface
End Class
