
#If Not %Def(%Const)
%Const = 1
Macro Const = Macro
Const LoadFactor = 0.75!
#EndIf

Function NewBag() Common As iBag
	Local vObj As iBag
	vObj = Class "cBag"
	Function = vObj
End Function

Class cBag Guid$("{B63899CE-36CF-4F33-A9E4-BA2557E77DD0}") Common
	Instance vBagEntries() As iBagEntry
	Instance vBuckets() As Long
	Instance vCount As Long
	Instance vCapacity As Long
	Instance vFull As Long

	Class Method Create()
		vFull = -1
		Me.Expand( 0 )
	End Method

	Class Method FindItem( ByRef pItem As iBagEntry ) As Long
			Local vBucketIndex As Long
			Local i As Long
			Local hc As Long
			Local be As iBagEntry

			hc = pItem.HashCode

			' Find entry
			vBucketIndex = hc Mod vCapacity
			i = vBuckets( vBucketIndex ) - 1
			While i >= 0 And IsObject( vBagEntries( i ) )
				be = vBagEntries( i )
				If be.HashCode = hc And pItem.Same( be ) Then
					Method = i
					Exit Method
				End If
				i = be.Next
			Wend

			Method = -1
	End Method

	' Adds a new string if it doesn't exist, either way returns the index of the item
	Class Method AddIfMissing( ByRef pItem As iBagEntry ) As Long
		Local vBucketIndex As Long
		Local i As Long

		i = Me.FindItem( pItem )
		If i>=0 Then
			Method = i
			Exit Method
		End If

		If vCount >= vFull Then
			Call Me.Expand( vCapacity + 1 )
		End If

		vBucketIndex = pItem.HashCode Mod vCapacity

		pItem.Next = vBuckets( vBucketIndex ) - 1
		vBagEntries( vCount ) = pItem
		Method = vCount
		Incr vCount
		vBuckets( vBucketIndex ) = vCount
	End Method

	' Expands the bag as needed
	Class Method Expand( ByVal capacity As Long )
		Local i As Long
		Local hc As Long
		Local vBucketIndex As Long
		Local pFull As Long


		capacity = BagGetCollectionSize( capacity )

		ReDim vBuckets( capacity-1 )
		pFull = CLng( capacity * LoadFactor )
		ReDim Preserve vBagEntries( pFull )

		' Redistribute hashes
		For i = 0 To vCount - 1
			vBucketIndex = vBagEntries( i ).HashCode Mod capacity

			vBagEntries( i ).Next = vBuckets( vBucketIndex ) - 1
			vBuckets( vBucketIndex ) = i + 1
		Next
		vCapacity = capacity
		vFull = pFull
	End Method

	Interface iBag Guid$("{DE5F2295-C83D-42A7-BB0C-48E44E99D360}")
		Inherit IUnknown

		' Returns index of added item
		Method Add( ByRef pItem As iBagEntry ) As Long
			Method = Me.AddIfMissing( pItem )
		End Method

		' Empty the bag
		Method Clear()
			Local i As Long
			For i=0 To vCount-1
				vBagEntries(i) = Nothing
			Next
			Reset vBuckets()

			vCount = 0
		End Method

		' -1 if found, 0 if not found
		Property Get Contains( ByRef pItem As iBagEntry ) As Long
			Property = Me.IndexOf( pItem ) <> -1
		End Property

		' Number of items in the bag
		Property Get Count() As Long
			Property = vCount
		End Property


		' Index of item if found, -1 if not found
		Method IndexOf( ByRef pItem As iBagEntry ) As Long
			Method = Me.FindItem( pItem )
		End Method

		Method Remove( ByRef pItem As iBagEntry )
			Me.RemoveAt( Me.FindItem( pItem ) )
		End Method

		Method RemoveAt( ByVal i As Long )
			Local vHashCode As Long
			Local vBucketIndex As Long
			Local j As Long
			Local k As Long

			' This is generally super expensive, use with caution
			If i<0 Or i>= vCount Then Exit Method

			vHashCode = vBagEntries(i).HashCode
			vBucketIndex = vHashCode Mod vCapacity

			If vBuckets( vBucketIndex ) - 1 = i Then
				vBuckets( vBucketIndex ) = vBagEntries(i).Next + 1: ' First item in list
			Else
				' Have prior entry's next point past the entry to remove
				j = vBuckets( vBucketIndex ) - 1
				Do While j>=0
					If vBagEntries( j ).Next = i Then
						vBagEntries( j ).Next = vBagEntries( i ).Next
						Exit Loop
					End If
					j = vBagEntries( j ).Next
				Loop
			End If
			' Ok now remove vBagEntries( i )
			vBagEntries( i ) = Nothing
			' Shift rest of the objects back (ouch exponential)
			For j = i To vCount-2
				vBagEntries( j ) = vBagEntries( j+1 )
				' Fix up Next references
				If vBagEntries( j ).Next > i Then
					vBagEntries( j ).Next = vBagEntries( j ).Next - 1
				End If
			Next
			Decr vCount
			' Release extra reference
			vBagEntries( vCount ) = Nothing
			' Adjust the bucket entries from i to the end, down by one
			If i > vCount-1000 Then
				' This is faster if you're pretty close to the end, decent if you remove the last
				For j=i To vCount-1
					vHashCode = vBagEntries( j ).HashCode
					vBucketIndex = vHashCode Mod vCapacity
					If vBuckets( vBucketIndex ) - 1 = j+1 Then
						Decr vBuckets( vBucketIndex )
					End If
				Next
			Else
				' This is better if you are not close to the end
				For j=0 To vCapacity
					If vBuckets( j ) > i Then
						Decr vBuckets( j )
					End If
				Next
			End If
			vHashCode = 0
		End Method

		' Get an item by index, returns empty string if out of bounds
		Property Get Item( ByVal index As Long ) As iBagEntry
			If index>=0 And index<vCount Then
				Property = vBagEntries( index )
			End If
		End Property
	End Interface
End Class

Function BagGetCollectionSize( ByVal capacity As Long ) Common As Long
	Static vPrimes() As Long
	Static vInit As Long
	Local i As Long

	If Not vInit Then
		ReDim vPrimes(25)
		Array Assign vPrimes() = _
						53, 97, 193, 389, 769, 1543, 3079, 6151, 12289 _
						, 24593, 49157, 98317, 196613, 393241, 786433 _
						, 1572869, 3145739, 6291469, 12582917, 25165843 _
						, 50331653, 100663319, 201326611, 402653189 _
						, 805306457, 1610612741
		vInit = -1
	End If

	For i=0 To 24
		If vPrimes(i)>capacity Then
			Function = vPrimes(i)
			Exit Function
		End If
	Next

	Function = vPrimes(25)
End Function
