Function NewLongStateRenderMap() Common Export As iLongStateRenderMap
	Dim o As iLongStateRenderMap
	o = Class "cLongStateRenderMap"
	Function = o
End Function

Class cLongStateRenderMap Guid$("{f5365363-1963-4f3f-88f5-b0b94c81ad9a}") Common
	Instance bag_ As iBag
	Class Method Create()
		bag_ = NewBag()
	End Method

	Interface iLongStateRenderMap Guid$("{ed79f931-1b11-483d-b2eb-6bed5335509c}")
		Inherit IUnknown

		' Returns index of added item
		Method Add( ByVal pKey As Long, ByVal pItem As iRender ) As Long
			Dim be As iBagEntry
			be = NewLongStateRenderMapEntry( pKey, pItem )
			bag_.Add( be )
		End Method

		' Empty the bag
		Method Clear()
			bag_.Clear()
		End Method

		' -1 if found, 0 if not found
		Property Get Contains( ByRef pKey As Long ) As Long
			Dim be As iBagEntry
			Dim no As iRender
			be = NewLongStateRenderMapEntry( pKey, no )

			Property = bag_.Contains( be )
		End Property

		' Number of items in the bag
		Property Get Count() As Long
			Property = bag_.Count
		End Property


		' Index of item if found, -1 if not found
		Method IndexOf( ByRef pKey As Long ) As Long
			Dim be As iBagEntry
			Dim no As iRender
			be = NewLongStateRenderMapEntry( pKey, no )
			Method = bag_.IndexOf( be )
		End Method

		Method Remove( ByRef pKey As Long )
			Dim be As iBagEntry
			Dim no As iRender
			be = NewLongStateRenderMapEntry( pKey, no )
			bag_.Remove( be )
		End Method

		Method RemoveAt( ByVal i As Long )
			bag_.RemoveAt( i )
		End Method

		' Get an item by index, returns empty string if out of bounds
		Property Get Value( ByVal pKey As Long ) As iRender
			Dim be As iBagEntry
			Dim no As iRender
			Dim iv As iLongStateRenderPair
			Local idx As Long
			Property = no
			idx = Me.IndexOf( pKey )
			be = bag_.Item( idx )
			iv = be
			If IsObject( iv ) Then
				Property = iv.Value
			End If
		End Property
		Property Set Value( ByVal pKey As Long, ByRef value As iRender )
			Dim be As iBagEntry
			Dim iv As iLongStateRenderPair
			Local idx As Long
			idx = Me.IndexOf( pKey )
			be = bag_.Item( idx )
			iv = be
			If IsObject( iv ) Then
				iv.Value = value
			End If
		End Property
		Property Get Key( ByVal index As Long ) As Long
			Dim be As iBagEntry
			Dim iv As iLongStateRenderPair
			Property = 0
			be = bag_.Item( index )
			iv = be
			If IsObject( iv ) Then
				Property = iv.Key
			End If
		End Property
	End Interface
End Class

Function NewLongStateRenderMapEntry( ByRef key As Long, ByRef value As iRender ) As iBagEntry
	Local a As iBagEntry
	Local b As iLongStateRenderPair

	a = Class "cLongStateRenderMapEntry"
	b = a
	b.Key = key
	b.Value = value
	Function = a
End Function

'=================================================================
' LongStateRenderMap entries (internal implementation class)
'=================================================================
Class cLongStateRenderMapEntry Guid$("{9813eccf-f3d4-4ef9-94ea-9797f8572f36}") Common
	Instance vHashCode As Long
	Instance vNext As Long
	Instance vKey As Long
	Instance vValue As iRender

	Interface iLongStateRenderPair Guid$("{247a58d0-da36-4330-bc91-09efd6302963}")
		Inherit IUnknown

		Property Get Value() As iRender
			Property = vValue
		End Property
		Property Set Value( ByVal value As iRender )
			vValue = value
		End Property

		Property Get Key () As Long
			Property = vKey
		End Property
		Property Set Key ( ByVal pVal  As Long)
			Local vHashCode As Long
			Local p As Byte Ptr
			Local i As Long

			If vKey <> pVal Then
				vKey  = pVal

				vHashCode = 1

				p = VarPtr( vKey )
				For i=1 To SizeOf(vKey)
					vHashCode = 1610612741 * vHashCode + @p
					Incr p
				Next
				vHashCode And= &H7FFFFFFF
			End If
		End Property
	End Interface

	Interface iBagEntry Guid$("{88765CD3-3E71-41CC-B0F3-29C8A4CE395A}")
		Inherit IUnknown

		Method Same( ByVal value As iBagEntry ) As Long
			Local a As iLongStateRenderPair
			Local b As iLongStateRenderPair
			a = Me
			b = value
			If IsObject(a) And IsObject(b) Then
				Method = IIf&( a.Key = b.Key, 1, 0 )
			End If
		End Method

		Property Get HashCode () As Long
			Property = vHashCode
		End Property
		Property Set HashCode ( ByVal pHashCode  As Long)
			vHashCode  = pHashCode
		End Property

		Property Get Next () As Long
			Property = vNext
		End Property
		Property Set Next ( ByVal pNext  As Long)
			vNext  = pNext
		End Property
	End Interface
End Class
